# -*- coding: utf-8 -*-
"""T-matrix tools using the `treams` toolkit

Tools to convert T-matrices to GPM structures

Requires treams (`pip install treams`).

Beutel, D., Fernandez-Corbaton, I. & Rockstuhl, C.
**treams – a T-matrix-based scattering code for nanophotonics.**
Computer Physics Communications 297, 109076 (2024)
https://github.com/tfp-photonics/treams

"""
import warnings

import numpy as np
import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import to_np


def spherical_wave_source_treams(
    k0, l, m, pol, mat_env=1.0, r_spherical_wave_source=[0, 0, 0], l_max=2
):
    """E and H field of a spherical wave source

    Args:
        k0 (float): wavnumber (2*pi / lambda0, in nm^-1)
        l (int): degree
        m (int): order
        pol (int): treams polarization
        mat_env (float, material): treams material or float. Float is interpreted as environment's permittivity.
        r_spherical_wave_source (list, optional): Cartesian location of source in nm. Defaults to [0, 0, 0].
        l_max (int, optional): maximum degree of expansion. Defaults to 2.

    Returns:
        spherical wave source

    Raises:
        ValueError: if required package `treams` is not found

    """
    try:
        # ignore import warnings
        with warnings.catch_warnings():
            import treams
    except ModuleNotFoundError:
        print("Requires `treams`, install via `pip install treams`.")
        raise

    spw = treams.spherical_wave(
        l=l, m=m, pol=pol, k0=k0, material=mat_env, modetype="regular", poltype="parity"
    )

    local_swb = treams.SphericalWaveBasis.default(
        l_max, positions=l_max * [r_spherical_wave_source]
    )

    ex_sph = spw.expand.apply_left(local_swb)

    return ex_sph


def cylindrical_wave_source_tream(
    k0,
    m,
    pol,
    k_par=0,
    mat_env=1.0,
    r_cyl_wave_source=[0, 0, 0],
    m_max=2,
):
    """E and H field of a cylindrical wave source in the XY(!!) plane

    Caution: infinite axis is along Z (=treams default; torchgdm default is Y axis)

    Args:
        k0 (float): wavnumber (2*pi / lambda0, in nm^-1)
        m (int): order
        pol (int): treams polarization. 0: TE (E in xy plane). 1: TM (E along z).
        k_par (float): wavevector component along the infinite axis (torchgdm: y-axis).
        mat_env (float, material): treams material or float. Float is interpreted as environment's permittivity.
        r_cyl_wave_source (list, optional): Cartesian location of source in nm. Defaults to [0, 0, 0].
        m_max (int, optional): maximum degree of expansion. Defaults to 2.

    Returns:
        cylindrical wave source

    Raises:
        ValueError: if required package `treams` is not found

    """
    try:
        # ignore import warnings
        with warnings.catch_warnings():
            import treams
    except ModuleNotFoundError:
        print("Requires `treams`, install via `pip install treams`.")
        raise

    cw = treams.cylindrical_wave(
        m=m,
        pol=pol,
        k0=k0,
        kz=k_par,
        material=mat_env,
        modetype="regular",
        poltype="parity",
    )

    local_cwb = treams.CylindricalWaveBasis.default(
        kzs=k_par, mmax=m_max, positions=r_cyl_wave_source
    )

    ex_sph = cw.expand.apply_left(local_cwb)

    return ex_sph


def setup_spherical_waves_treams(r_source, k0, l_max, m_max, mat_env=1.0):
    try:
        # ignore import warnings
        with warnings.catch_warnings():
            import treams
    except ModuleNotFoundError:
        print("Requires `treams`, install via `pip install treams`.")
        raise

    r_source = np.atleast_2d(r_source)

    inc_fields = []
    for i, r_s in enumerate(r_source):

        # generate random configurations for each source
        _l = np.random.randint(1, l_max + 1)
        _pol = np.random.randint(0, 2)  # 0 or 1
        m_lim = min([_l, m_max])
        _m = np.random.randint(-m_lim, m_lim + 1)

        sw_source = spherical_wave_source_treams(
            k0=k0,
            l=_l,
            m=_m,
            pol=_pol,
            mat_env=mat_env,
            r_spherical_wave_source=r_s,
            l_max=_l,
        )
        inc_fields.append(sw_source)

    return inc_fields


def setup_cylindrical_waves_treams(r_source, k0, m_max, mat_env=1.0):
    try:
        # ignore import warnings
        with warnings.catch_warnings():
            import treams
    except ModuleNotFoundError:
        print("Requires `treams`, install via `pip install treams`.")
        raise

    r_source = np.atleast_2d(r_source)

    inc_fields = []
    for i, r_s in enumerate(r_source):

        # generate random configurations for each source
        _m = np.random.randint(-m_max, m_max + 1)
        _pol = np.random.randint(0, 2)  # 0 or 1
        _mmax = max([1, abs(_m)])
        cw_source = cylindrical_wave_source_tream(
            k0=k0, m=_m, pol=_pol, mat_env=mat_env, r_cyl_wave_source=r_s, m_max=_mmax
        )
        inc_fields.append(cw_source)

    return inc_fields


def setup_plane_waves_treams(
    n_angles, k0, mat_env, e0_amplitude=1.0, inc_planes=["xz", "xy"]
):
    try:
        # ignore import warnings
        with warnings.catch_warnings():
            import treams
    except ModuleNotFoundError:
        print("Requires `treams`, install via `pip install treams`.")
        raise

    _e0 = e0_amplitude
    inc_fields = []
    for angle in np.linspace(0, 2 * np.pi, n_angles, endpoint=False):

        # - XZ plane
        if "xz" in inc_planes:
            k0_vec = [np.sin(angle), 0, np.cos(angle)]
            # s-pol
            e0_vec = [0, _e0, 0]
            inc_fields.append(
                treams.plane_wave(
                    k0_vec, pol=e0_vec, k0=k0, material=mat_env, poltype="parity"
                )
            )
            # p-pol
            e0_vec = [np.cos(angle) * _e0, 0, np.sin(angle) * _e0]
            inc_fields.append(
                treams.plane_wave(
                    k0_vec, pol=e0_vec, k0=k0, material=mat_env, poltype="parity"
                )
            )

        # - YZ plane
        if "yz" in inc_planes:
            k0_vec = [0, np.sin(angle), np.cos(angle)]
            # s-pol
            e0_vec = [_e0, 0, 0]
            inc_fields.append(
                treams.plane_wave(
                    k0_vec, pol=e0_vec, k0=k0, material=mat_env, poltype="parity"
                )
            )
            # p-pol
            e0_vec = [0, np.cos(angle) * _e0, np.sin(angle) * _e0]
            inc_fields.append(
                treams.plane_wave(
                    k0_vec, pol=e0_vec, k0=k0, material=mat_env, poltype="parity"
                )
            )

        # - XY plane
        if "xy" in inc_planes:
            k0_vec = [np.sin(angle), np.cos(angle), 0]
            # s-pol
            e0_vec = [0, 0, _e0]
            inc_fields.append(
                treams.plane_wave(
                    k0_vec, pol=e0_vec, k0=k0, material=mat_env, poltype="parity"
                )
            )
            # p-pol
            e0_vec = [np.cos(angle) * _e0, np.sin(angle) * _e0, 0]
            inc_fields.append(
                treams.plane_wave(
                    k0_vec, pol=e0_vec, k0=k0, material=mat_env, poltype="parity"
                )
            )

    return inc_fields


# - parallelized treams evaluation (illumination and scattering)
def _eval_treams(inc, tmatrix, r_probe, r_gpm):
    try:
        # ignore import warnings
        with warnings.catch_warnings():
            import treams
    except ModuleNotFoundError:
        print("Requires `treams`, install via `pip install treams`.")
        raise

    if type(tmatrix) == treams.TMatrixC:
        assert torch.all(r_probe[..., 2] == 0), "2D treams: all z pos must be 0!"
        assert torch.all(r_gpm[..., 2] == 0), "2D treams: all z pos must be 0!"

    # - parity polarization
    if tmatrix.poltype != "parity":
        tm = tmatrix.changepoltype("parity")
    else:
        tm = tmatrix
    sca = tm @ inc.expand(tm.basis)

    # - calc incident and scattered fields
    _e_sca = sca.efield(to_np(r_probe))._array
    _h_sca = sca.hfield(to_np(r_probe))._array
    _e_inc = inc.efield(to_np(r_gpm))._array
    _h_inc = inc.hfield(to_np(r_gpm))._array

    return _e_sca, _h_sca, _e_inc, _h_inc


def convert_tmatrix3D_to_GPM(
    tmatrix,
    r_gpm,
    r_probe,
    n_src_pw_angles,
    n_src_local,
    radius_src_local,
    l_max_sw_inc=2,
    m_max_sw_inc=2,
    normalize_inc_fields=True,
    n_proc=None,
    **kwargs,
):
    try:
        from joblib import Parallel, delayed
        from joblib import cpu_count

        if n_proc is None:
            n_proc = cpu_count(only_physical_cores=True)
    except ModuleNotFoundError:
        print("Requires `joblib`, install via `pip install joblib`.")
        raise

    try:
        # ignore import warnings
        with warnings.catch_warnings():
            import treams
    except ModuleNotFoundError:
        print("Requires `treams`, install via `pip install treams`.")
        raise

    from torchgdm.env import EnvHomogeneous3D
    from torchgdm.struct.eff_model_tools import extract_gpm_from_fields
    from torchgdm.tools.geometry import sample_random_spherical
    from torchgdm.tools.misc import to_np

    _device_int = "cpu"  # torchgdm part on CPU
    assert type(tmatrix) == treams.TMatrix, "Requires treams 3D T-Matrix"
    assert tmatrix.material.epsilon.imag == 0, "absorbing environments not supported"

    _env_3d = EnvHomogeneous3D(env_material=float(tmatrix.material.epsilon.real))

    # --- setup illumination sources
    # - plane waves
    inc_fields_pw = setup_plane_waves_treams(
        n_src_pw_angles,
        k0=tmatrix.k0,
        mat_env=tmatrix.material,
        inc_planes=["xz", "yz", "xy"],
    )

    # - spherical waves
    if n_src_local > 0:
        r_sw_src = sample_random_spherical(n_src_local) * radius_src_local
        inc_fields_dp = setup_spherical_waves_treams(
            r_source=to_np(r_sw_src),
            k0=tmatrix.k0,
            l_max=l_max_sw_inc,
            m_max=m_max_sw_inc,
            mat_env=tmatrix.material,
        )
    else:
        inc_fields_dp = []
    inc_fields = inc_fields_pw + inc_fields_dp

    # - scattering and illuminations
    # illum_results = [_eval_treams(inc, tmatrix, r_probe, r_gpm) for inc in inc_fields]
    illum_results = Parallel(n_jobs=n_proc)(
        delayed(_eval_treams)(inc, tmatrix, r_probe, r_gpm) for inc in inc_fields
    )

    # - fill full arrays
    e_inc_treams = np.zeros((len(inc_fields), len(r_gpm), 3), dtype=np.complex128)
    h_inc_treams = np.zeros((len(inc_fields), len(r_gpm), 3), dtype=np.complex128)
    e_sca_treams = np.zeros((len(inc_fields), len(r_probe), 3), dtype=np.complex128)
    h_sca_treams = np.zeros((len(inc_fields), len(r_probe), 3), dtype=np.complex128)
    for i_inc, fields in enumerate(illum_results):
        if normalize_inc_fields:
            norm = max([np.abs(_f).max() for _f in fields])
        else:
            norm = 1.0

        e_sca_treams[i_inc] = fields[0] / norm
        h_sca_treams[i_inc] = fields[1] / norm
        e_inc_treams[i_inc] = fields[2] / norm
        h_inc_treams[i_inc] = fields[3] / norm

    # - optimize GPM
    GPM_dict = extract_gpm_from_fields(
        wavelength=2 * np.pi / tmatrix.k0,
        efields_sca=torch.as_tensor(e_sca_treams, device=_device_int).to(
            dtype=DTYPE_COMPLEX
        ),
        hfields_sca=torch.as_tensor(h_sca_treams, device=_device_int).to(
            dtype=DTYPE_COMPLEX
        ),
        efields_inc=torch.as_tensor(e_inc_treams, device=_device_int).to(
            dtype=DTYPE_COMPLEX
        ),
        hfields_inc=torch.as_tensor(h_inc_treams, device=_device_int).to(
            dtype=DTYPE_COMPLEX
        ),
        r_probe=r_probe,
        r_gpm=r_gpm,
        environment=_env_3d,
        device=_device_int,
        verbose=False,
        return_all_results=True,
        **kwargs,
    )

    return GPM_dict


def convert_tmatrix2D_to_GPM(
    tmatrix,
    r_gpm,
    r_probe,
    n_src_pw_angles,
    n_src_local,
    radius_src_local,
    m_max_sw_inc=1,
    normalize_inc_fields=True,
    n_proc=None,
    **kwargs,
):
    try:
        from joblib import Parallel, delayed
        from joblib import cpu_count

        if n_proc is None:
            n_proc = cpu_count(only_physical_cores=True)
    except ModuleNotFoundError:
        print("Requires `joblib`, install via `pip install joblib`.")
        raise

    try:
        # ignore import warnings
        with warnings.catch_warnings():
            import treams
    except ModuleNotFoundError:
        print("Requires `treams`, install via `pip install treams`.")
        raise

    from torchgdm.env import EnvHomogeneous2D
    from torchgdm.struct.eff_model_tools import extract_gpm_from_fields
    from torchgdm.tools.geometry import sample_random_circular
    from torchgdm.tools.misc import to_np

    _device_int = "cpu"  # torchgdm part on CPU
    assert type(tmatrix) == treams.TMatrixC, "Requires treams 2D T-Matrix (`TMatrixC`)"
    assert tmatrix.material.epsilon.imag == 0, "absorbing environments not supported"

    _env_2d = EnvHomogeneous2D(env_material=float(tmatrix.material.epsilon.real))

    # --- setup illumination sources
    # - plane waves perp. to infinite axis
    inc_fields_pw = setup_plane_waves_treams(
        n_src_pw_angles, k0=tmatrix.k0, mat_env=tmatrix.material, inc_planes=["xy"]
    )

    # - cylindrical waves
    if n_src_local > 0:
        r_sw_src = (
            sample_random_circular(n_src_local, projection="xy") * radius_src_local
        )
        from torchgdm.tools.geometry import coordinate_map_1d_circular

        r_sw_src = coordinate_map_1d_circular(
            r=radius_src_local,
            n_phi=n_src_local,
            device=_device_int,
        )["r_probe"]
        r_sw_src = r_sw_src[..., [0, 2, 1]]  # XY plane
        inc_fields_dp = setup_cylindrical_waves_treams(
            r_source=to_np(r_sw_src),
            k0=tmatrix.k0,
            m_max=m_max_sw_inc,
            mat_env=tmatrix.material,
        )
    else:
        inc_fields_dp = []
    inc_fields = inc_fields_pw + inc_fields_dp

    # - scattering and illuminations
    r_probe_z_infinite = r_probe.clone()
    r_probe_z_infinite = r_probe_z_infinite[:, [0, 2, 1]]
    r_gpm_z_infinite = r_gpm.clone()
    r_gpm_z_infinite = r_gpm_z_infinite[:, [0, 2, 1]]
    # illum_results = [_eval_treams(inc, tmatrix, r_probe_z_infinite, r_gpm_z_infinite) for inc in inc_fields]
    illum_results = Parallel(n_jobs=n_proc)(
        delayed(_eval_treams)(inc, tmatrix, r_probe_z_infinite, r_gpm_z_infinite)
        for inc in inc_fields
    )

    # - fill full arrays
    e_inc_treams = np.zeros((len(inc_fields), len(r_gpm), 3), dtype=np.complex128)
    h_inc_treams = np.zeros((len(inc_fields), len(r_gpm), 3), dtype=np.complex128)
    e_sca_treams = np.zeros((len(inc_fields), len(r_probe), 3), dtype=np.complex128)
    h_sca_treams = np.zeros((len(inc_fields), len(r_probe), 3), dtype=np.complex128)

    idx_treams = [0, 2, 1]  # treams is infinite along z
    for i_inc, fields in enumerate(illum_results):
        if normalize_inc_fields:
            norm = max([np.abs(_f).max() for _f in fields])
        else:
            norm = 1.0

        e_sca_treams[i_inc] = fields[0][..., idx_treams] / norm
        h_sca_treams[i_inc] = fields[1][..., idx_treams] / norm
        e_inc_treams[i_inc] = fields[2][..., idx_treams] / norm
        h_inc_treams[i_inc] = fields[3][..., idx_treams] / norm

    # - adapt sign of y-comp. (2D torchgdm infinite axis) component
    # treams infinite axis opposite handedness convention
    field_directions = [1, -1, 1]
    for i_f, sign in enumerate(field_directions):
        e_sca_treams[..., i_f] = e_sca_treams[..., i_f] * sign
        h_sca_treams[..., i_f] = h_sca_treams[..., i_f] * sign
        e_inc_treams[..., i_f] = e_inc_treams[..., i_f] * sign
        h_inc_treams[..., i_f] = h_inc_treams[..., i_f] * sign

    # - optimize GPM
    GPM_dict = extract_gpm_from_fields(
        wavelength=2 * np.pi / tmatrix.k0,
        efields_sca=torch.as_tensor(e_sca_treams, device=_device_int).to(
            dtype=DTYPE_COMPLEX
        ),
        hfields_sca=torch.as_tensor(h_sca_treams, device=_device_int).to(
            dtype=DTYPE_COMPLEX
        ),
        efields_inc=torch.as_tensor(e_inc_treams, device=_device_int).to(
            dtype=DTYPE_COMPLEX
        ),
        hfields_inc=torch.as_tensor(h_inc_treams, device=_device_int).to(
            dtype=DTYPE_COMPLEX
        ),
        r_probe=r_probe,
        r_gpm=r_gpm,
        environment=_env_2d,
        device=_device_int,
        verbose=False,
        return_all_results=True,
        **kwargs,
    )

    return GPM_dict


def _test_effective_model_accuracy_3d(
    struct_alpha,
    tm_list,
    which=["ecs", "nf_tot"],
    environment=None,
    r_enclosing=None,
    rtol=0.10,
    verbose=True,
    progress_bar=False,
    device=None,
):
    """test effective torchgdm model against a T-Matrix simulation (3D)

    Print some information about model accuracy

    Args:
        struct_alpha (torchgdm.struct3d.StructEffPola3D): effective polarizability structure
        tm_list (list of treams TMatrix): list of reference T-Matrix for each wavelength
        which_test (list of str, optional): Which observables to test. Can contain any of ["ecs", "nf_sca", "nf_tot"]. Defaults to ["ecs", "nf_tot"].
        environment (3D env. class, optional): Simulation environment. If None, use environment from first T-Matrix. Defaults to None.
        r_enclosing (float, optional): T-Matrix enclosing sphere radius, used for nearfield check. If not given, guess some value from the GPM. Defaults to None.
        rtol (float, optional): relative error threshold for raising warnings. Defaults to 0.10.
        verbose (bool, optional): Print detailed info. Defaults to True.
        progress_bar (bool, optional): Show progress bars. Defaults to False.
        device (str, optional): If None, use structure's device. Defaults to None.

    """
    try:
        # ignore import warnings
        with warnings.catch_warnings():
            import treams
    except ModuleNotFoundError:
        print("Requires `treams`, install via `pip install treams`.")
        raise

    from torchgdm.tools.batch import calc_spectrum
    from torchgdm.postproc import crosssect
    from torchgdm.simulation import Simulation
    from torchgdm.tools.geometry import coordinate_map_1d
    from torchgdm.tools.misc import to_np
    from torchgdm.tools.misc import _check_environment

    if device is None:
        device = struct_alpha.device

    # environment
    n_dim = struct_alpha.n_dim
    if n_dim == 3:
        from torchgdm.env.freespace_3d.inc_fields import PlaneWave
    else:
        raise ValueError(
            f"This test is for 3D structures. Given structure has N_dim={n_dim}"
        )

    if environment is None:  # use first T-Matrix
        environment = complex(tm_list[0].material.epsilon)
    env = _check_environment(environment, N_dim=n_dim, device=device)

    wavelengths = struct_alpha.wavelengths_data

    # guess some circumscribing radius if not given
    if r_enclosing is None:
        r_enclosing = (
            struct_alpha.get_all_positions().max() * 2 + struct_alpha.step.max() * 2
        )

    # nearfield calc. positions
    max_step = struct_alpha.step.max()
    z_max = r_enclosing + max_step * 2.0
    r_probe = coordinate_map_1d(
        [-500, 500], 10, r2=50, r3=z_max, direction="x", device=device
    )["r_probe"]

    # - treams
    cs_tm = {}
    cs_tm["scs"] = []
    cs_tm["ecs"] = []
    nf_tm = {}
    nf_tm["inc"] = []
    nf_tm["sca"] = []
    nf_tm["tot"] = []
    for i_wl, wl in enumerate(wavelengths):
        k0 = float(2 * np.pi / wl)
        eps_env = to_np(env.env_material.get_epsilon(wavelength=wl).real)[0, 0]
        tm = tm_list[i_wl]
        if tm.poltype != "parity":
            tm = tm.changepoltype("parity")

        kw_inc = dict(k0=k0, material=eps_env, poltype="parity")
        inc_list = [
            treams.plane_wave(kvec=[0, 0, 1], pol=[-1, 0, 0], **kw_inc),
            treams.plane_wave(kvec=[0, 0, 1], pol=[0, 1, 0], **kw_inc),
            treams.plane_wave(kvec=[-1, 0, 0], pol=[0, 0, -1], **kw_inc),
            treams.plane_wave(kvec=[-1, 0, 0], pol=[0, 1, 0], **kw_inc),
            treams.plane_wave(kvec=[0, -1, 0], pol=[0, 0, -1], **kw_inc),
            treams.plane_wave(kvec=[0, -1, 0], pol=[-1, 0, 0], **kw_inc),
        ]

        cs_tm["scs"].append([tm.xs(inc)[0] for inc in inc_list])
        cs_tm["ecs"].append([tm.xs(inc)[1] for inc in inc_list])

        for inc in inc_list:
            sca = tm @ inc.expand(tm.basis)
            _ef_tm = sca.efield(r_probe)
            _ef_inc = inc.efield(r_probe)
            nf_tm["inc"].append(_ef_inc)
            nf_tm["sca"].append(_ef_tm)
            nf_tm["tot"].append(_ef_inc + _ef_tm)

    cs_tm["scs"] = np.array(cs_tm["scs"])
    cs_tm["ecs"] = np.array(cs_tm["ecs"])

    nf_tm["inc"] = np.array(nf_tm["inc"])
    nf_tm["sca"] = np.array(nf_tm["sca"])
    nf_tm["tot"] = np.array(nf_tm["tot"])

    e_inc_list = [
        PlaneWave(e0p=1.0, e0s=0.0, inc_angle=0),  # k||z
        PlaneWave(e0p=0.0, e0s=1.0, inc_angle=0),  # k||z
        PlaneWave(e0p=1.0, e0s=0.0, inc_angle=torch.pi / 2),  # k||x
        PlaneWave(e0p=0.0, e0s=1.0, inc_angle=torch.pi / 2),  # k||x
        PlaneWave(e0p=1.0, e0s=0.0, inc_plane="yz", inc_angle=torch.pi / 2),  # k||y
        PlaneWave(e0p=0.0, e0s=1.0, inc_plane="yz", inc_angle=torch.pi / 2),  # k||y
    ]

    # setup a discretized and a effective dipole pair simulation
    sim_alpha = Simulation(
        structures=[struct_alpha],
        environment=env,
        illumination_fields=e_inc_list,
        wavelengths=wavelengths,
        device=device,
    )

    # run simulations and calc. cross section spectra
    if verbose:
        print("-" * 60)
        print("Testing effective model vs. T-Matrix (treams, 3D).")

    sim_alpha.run(verbose=False, progress_bar=progress_bar)

    test_results = dict()

    # calculate cross section errors
    cs_alpha = sim_alpha.get_spectra_crosssections(progress_bar=progress_bar)

    which_cs = []
    if "scs" in which:
        which_cs.append("scs")
    if "ecs" in which:
        which_cs.append("ecs")

    for k in which_cs:
        try:
            rel_diff = (cs_tm[k] - to_np(cs_alpha[k])) / (
                (cs_tm[k] + to_np(cs_alpha[k]))
            )
            test_results[k] = rel_diff

            mean_rel_error = np.mean(np.abs(rel_diff))
            peak_rel_error = np.max(np.abs(rel_diff))
            if verbose:
                print("cross sections - '{}':".format(k))
                print("    - mean rel. error: {:.3g}".format(mean_rel_error))
                print("    - peak rel. error: {:.3g}".format(peak_rel_error))
            if mean_rel_error > rtol:
                warnings.warn(
                    "'{}': Effective model mean relative error larger than tolerance ({:.3f} > {:.3f})".format(
                        k, mean_rel_error, rtol
                    )
                )

            elif peak_rel_error > rtol:
                warnings.warn(
                    "'{}': Effective model peak relative error larger than tolerance ({:.3f} > {:.3f})".format(
                        k, peak_rel_error, rtol
                    )
                )

            # print(cs_tm[k], rel_diff)
        except TypeError:
            pass

    # calculate nearfield errors several steps above structure
    nf_alpha = sim_alpha.get_spectra_nf(r_probe=r_probe, progress_bar=progress_bar)

    which_nf = []
    if "nf_tot" in which:
        which_nf.append("tot")
    if "nf_sca" in which:
        which_nf.append("sca")
    for k in which_nf:
        try:
            e_alpha = np.array([to_np(_nf.efield) for _nf in nf_alpha[k]])
            e_alpha = e_alpha.flatten()
            e_discr = to_np(nf_tm[k])
            e_discr = e_discr.flatten()
            rel_diff = (e_discr - e_alpha) / ((e_discr.max() + e_alpha.max()))
            test_results[k] = rel_diff

            mean_rel_error = np.mean(np.abs(rel_diff))
            peak_rel_error = np.max(np.abs(rel_diff))
            if verbose:
                print("nearfield - '{}' (at z={}nm):".format(k, int(to_np(z_max))))
                print("    - mean rel. error: {:.3g}".format(to_np(mean_rel_error)))
                print("    - peak rel. error: {:.3g}".format(to_np(peak_rel_error)))
            if mean_rel_error > rtol:
                warnings.warn(
                    "'{}': Effective model mean relative error larger than tolerance ({:.3f} > {:.3f})".format(
                        k, mean_rel_error, rtol
                    )
                )

            elif peak_rel_error > rtol:
                warnings.warn(
                    "'{}': Effective model peak relative error larger than tolerance ({:.3f} > {:.3f})".format(
                        k, peak_rel_error, rtol
                    )
                )
            # print('avg err per field config', np.mean(rel_diff, axis=(2)))

        except TypeError:
            pass
    if verbose:
        print("-" * 60)

    return test_results


def _test_effective_model_accuracy_2d(
    struct_alpha,
    tm_list,
    which=["ecs", "nf_tot"],
    environment=None,
    r_enclosing=None,
    rtol=0.10,
    verbose=True,
    progress_bar=False,
    device=None,
):
    """test effective torchgdm model against a T-Matrix simulation (2D)

    Print some information about model accuracy

    Args:
        struct_alpha (torchgdm.struct3d.StructEffPola3D): effective polarizability structure
        tm_list (list of treams TMatrix): list of reference T-Matrix for each wavelength
        which_test (list of str, optional): Which observables to test. Can contain any of ["ecs", "nf_sca", "nf_tot"]. Defaults to ["ecs", "nf_tot"].
        environment (2D env. class, optional): Simulation environment. If None, use environment from first T-Matrix. Defaults to None.
        r_enclosing (float, optional): T-Matrix enclosing sphere radius, used for nearfield check. If not given, guess some value from the GPM. Defaults to None.
        rtol (float, optional): relative error threshold for raising warnings. Defaults to 0.10.
        verbose (bool, optional): Print detailed info. Defaults to True.
        progress_bar (bool, optional): Show progress bars. Defaults to False.
        device (str, optional): If None, use structure's device. Defaults to None.

    """
    try:
        # ignore import warnings
        with warnings.catch_warnings():
            import treams
    except ModuleNotFoundError:
        print("Requires `treams`, install via `pip install treams`.")
        raise

    from torchgdm.tools.batch import calc_spectrum
    from torchgdm.simulation import Simulation
    from torchgdm.tools.misc import to_np
    from torchgdm.tools.misc import _check_environment
    from torchgdm.tools.geometry import coordinate_map_1d

    if device is None:
        device = struct_alpha.device

    # environment
    n_dim = struct_alpha.n_dim
    if environment is None:  # use first T-Matrix
        environment = complex(tm_list[0].material.epsilon)
    env = _check_environment(environment, N_dim=n_dim, device=device)

    wavelengths = struct_alpha.wavelengths_data

    # guess some circumscribing radius if not given
    if r_enclosing is None:
        r_enclosing = (
            struct_alpha.get_all_positions().max() * 2 + struct_alpha.step.max() * 2
        )

    # nearfield calc. positions
    max_step = struct_alpha.step.max()
    z_max = r_enclosing + max_step * 2.0
    r_probe = coordinate_map_1d(
        [-500, 500], 40, r2=0, r3=z_max, direction="x", device=device
    )["r_probe"]

    # probe pos. in treams coordinate convention (Y <--> Z)
    r_prb_tr = to_np(r_probe).copy()
    r_prb_tr = r_prb_tr[:, [0, 2, 1]]

    if n_dim == 2:
        from torchgdm.env.freespace_2d.inc_fields import PlaneWave
    else:
        raise ValueError(
            f"This test is for 3D structures. Given structure has N_dim={n_dim}"
        )

    # - treams
    cs_tm = {}
    cs_tm["scs"] = []
    cs_tm["ecs"] = []
    nf_tm = {}
    nf_tm["inc"] = []
    nf_tm["sca"] = []
    nf_tm["tot"] = []
    for i_wl, wl in enumerate(wavelengths):
        k0 = float(2 * np.pi / wl)
        eps_env = to_np(env.env_material.get_epsilon(wavelength=wl).real)[0, 0]
        tm = tm_list[i_wl]
        if tm.poltype != "parity":
            tm = tm.changepoltype("parity")

        # 2D treams: y <--> z; treams z-comp: sign
        kw_inc = dict(k0=k0, material=eps_env, poltype="parity")
        # inc_list = [treams.plane_wave(
        #     kvec=[0, 1, 0], pol=[1,0,0], k0=tm.k0, material=tm.material, poltype="parity"
        # )]
        inc_list = [
            treams.plane_wave(kvec=[0, 1, 0], pol=[-1, 0, 0], **kw_inc),
            treams.plane_wave(kvec=[0, 1, 0], pol=[0, 0, -1], **kw_inc),
            treams.plane_wave(kvec=[-1, 0, 0], pol=[0, -1, 0], **kw_inc),
            treams.plane_wave(kvec=[-1, 0, 0], pol=[0, 0, -1], **kw_inc),
        ]
        _scs_ecs = [tm.xw(inc) for inc in inc_list]
        cs_tm["scs"].append(_scs_ecs)
        cs_tm["ecs"].append(_scs_ecs)

        for inc in inc_list:
            sca = tm @ inc.expand(tm.basis)
            _ef_tm = sca.efield(r_prb_tr)
            _ef_inc = inc.efield(r_prb_tr)
            nf_tm["inc"].append(_ef_inc)
            nf_tm["sca"].append(_ef_tm)
            nf_tm["tot"].append(_ef_inc + _ef_tm)

    cs_tm["scs"] = np.array(cs_tm["scs"])[..., 0]  # keep only scs
    cs_tm["ecs"] = np.array(cs_tm["ecs"])[..., 1]  # keep only ecs
    # adapt 2D treams to torchgdm conventions
    nf_tm["inc"] = np.array(nf_tm["inc"])[..., [0, 2, 1]]
    nf_tm["sca"] = np.array(nf_tm["sca"])[..., [0, 2, 1]]
    nf_tm["tot"] = np.array(nf_tm["tot"])[..., [0, 2, 1]]
    nf_tm["inc"][..., 1] = -1 * nf_tm["inc"][..., 1]
    nf_tm["sca"][..., 1] = -1 * nf_tm["sca"][..., 1]
    nf_tm["tot"][..., 1] = -1 * nf_tm["tot"][..., 1]

    e_inc_list = [
        PlaneWave(e0p=1.0, e0s=0.0, inc_angle=0),  # k||z
        PlaneWave(e0p=0.0, e0s=1.0, inc_angle=0),  # k||z
        PlaneWave(e0p=1.0, e0s=0.0, inc_angle=torch.pi / 2),  # k||x
        PlaneWave(e0p=0.0, e0s=1.0, inc_angle=torch.pi / 2),  # k||x
    ]

    # setup a discretized and a effective dipole pair simulation
    sim_alpha = Simulation(
        structures=[struct_alpha],
        environment=env,
        illumination_fields=e_inc_list,
        wavelengths=wavelengths,
        device=device,
    )

    # run simulations and calc. cross section spectra
    if verbose:
        print("-" * 60)
        print("Testing effective model vs. T-Matrix (treams, 2D).")

    sim_alpha.run(verbose=False, progress_bar=progress_bar)

    test_results = dict()

    # calculate cross section errors
    cs_alpha = sim_alpha.get_spectra_crosssections(progress_bar=progress_bar)

    which_cs = []
    if "scs" in which:
        which_cs.append("scs")
    if "ecs" in which:
        which_cs.append("ecs")
    for k in which_cs:
        try:
            rel_diff = (cs_tm[k] - to_np(cs_alpha[k])) / (
                (cs_tm[k] + to_np(cs_alpha[k]))
            )
            test_results[k] = rel_diff

            mean_rel_error = np.mean(np.abs(rel_diff))
            peak_rel_error = np.max(np.abs(rel_diff))
            if verbose:
                print("cross sections - '{}':".format(k))
                print("    - mean rel. error: {:.3g}".format(mean_rel_error))
                print("    - peak rel. error: {:.3g}".format(peak_rel_error))
            if mean_rel_error > rtol:
                warnings.warn(
                    "'{}': Effective model mean relative error larger than tolerance ({:.3f} > {:.3f})".format(
                        k, mean_rel_error, rtol
                    )
                )

            elif peak_rel_error > rtol:
                warnings.warn(
                    "'{}': Effective model peak relative error larger than tolerance ({:.3f} > {:.3f})".format(
                        k, peak_rel_error, rtol
                    )
                )
            # print(rel_diff)
            # print(cs_tm[k])
            # print(to_np(cs_alpha[k]))
        except TypeError:
            pass

    # calculate nearfield errors several steps above structure
    nf_alpha = sim_alpha.get_spectra_nf(r_probe=r_probe, progress_bar=progress_bar)

    which_nf = []
    if "nf_tot" in which:
        which_nf.append("tot")
    if "nf_sca" in which:
        which_nf.append("sca")
    for k in which_nf:
        try:
            e_alpha = np.array([to_np(_nf.efield) for _nf in nf_alpha[k]])
            e_alpha = e_alpha.flatten()
            e_discr = to_np(nf_tm[k])
            e_discr = e_discr.flatten()
            rel_diff = (e_discr - e_alpha) / ((e_discr.max() + e_alpha.max()))
            test_results[k] = rel_diff

            mean_rel_error = np.mean(np.abs(rel_diff))
            peak_rel_error = np.max(np.abs(rel_diff))
            if verbose:
                print("nearfield - '{}' (at z={}nm):".format(k, int(to_np(z_max))))
                print("    - mean rel. error: {:.3g}".format(to_np(mean_rel_error)))
                print("    - peak rel. error: {:.3g}".format(to_np(peak_rel_error)))
            if mean_rel_error > rtol:
                warnings.warn(
                    "'{}': Effective model mean relative error larger than tolerance ({:.3f} > {:.3f})".format(
                        k, mean_rel_error, rtol
                    )
                )

            elif peak_rel_error > rtol:
                warnings.warn(
                    "'{}': Effective model peak relative error larger than tolerance ({:.3f} > {:.3f})".format(
                        k, peak_rel_error, rtol
                    )
                )
            # print('avg err per field config', np.mean(rel_diff, axis=(2)))

        except TypeError:
            pass
    if verbose:
        print("-" * 60)

    return test_results



# -*- coding: utf-8 -*-
"""2D surface discretizations

.. currentmodule:: torchgdm.struct.struct2d

Classes
-------

.. autosummary::
   :toctree: generated/

   StructDiscretized2D
   StructDiscretizedSquare2D
   StructGPM2D
   StructMieCylinderGPM2D
   StructTMatrixGPM2D


dipolar approximation
^^^^^^^^^^^^^^^^^^^^^

.. autosummary::
   :toctree: generated/
   
   StructEffPola2D
   StructMieCylinderEffPola2D
   

Functions
---------

.. autosummary::
   :toctree: generated/

   extract_eff_pola_2d
   extract_gpm_from_struct
   get_gpm_positions_by_clustering


Geometries
----------

.. autosummary::
   :toctree: generated/

   square
   rectangle
   circle
   split_ring
   triangle_equilateral


Discretizer functions
---------------------

.. autosummary::
   :toctree: generated/

   discretizer_square

"""
from . import surface
from . import line
from . import gpm2d
from . import geometries

from .surface import StructDiscretized2D
from .surface import StructDiscretizedSquare2D
from .gpm2d import StructGPM2D
from .gpm2d import StructMieCylinderGPM2D
from .gpm2d import StructTMatrixGPM2D
from .line import StructEffPola2D
from .line import StructMieCylinderEffPola2D

# from .line import extract_eff_pola_2d
from torchgdm.struct.eff_model_tools import extract_eff_pola_2d
from torchgdm.struct.eff_model_tools import extract_gpm_from_struct
from torchgdm.struct.eff_model_tools import get_gpm_positions_by_clustering

from .geometries import discretizer_square

from .geometries import square
from .geometries import rectangle
from .geometries import circle
from .geometries import split_ring
from .geometries import triangle_equilateral



# -*- coding: utf-8 -*-
"""2D surface discretization structure classes"""
import warnings

import torch

from torchgdm.constants import (
    DTYPE_FLOAT,
    DTYPE_COMPLEX,
    COLORS_DEFAULT,
)
from torchgdm.tools.misc import get_default_device
from torchgdm.struct.base_classes import StructBase
from torchgdm.tools.geometry import get_step_from_geometry
from torchgdm.tools.geometry import rotation_y
from torchgdm.tools.misc import ptp


class StructDiscretized2D(StructBase):
    """base class 2D surface discretized structure (infinite y axis)

    Using a list of positions in the XZ-plane and materials,
    this class defines the basic 2D surface discretization, the
    polarizabilities and self-terms
    """

    __name__ = "2D discretized structure"

    def __init__(
        self,
        positions: torch.Tensor,
        materials,
        step=None,
        on_distance_violation: str = "warn",
        radiative_correction: bool = False,
        device: torch.device = None,
        **kwargs,
    ):
        """2D discretized structure

        Args:
            positions (torch.Tensor): meshpoint positions (3D, but all y values must be zero)
            materials (material class, or list of them): material of structure. If list, define the material of each meshpoint.
            step (float, optional): nominal stepsize. If not provided, infer automatically from geometry. Defaults to None.
            on_distance_violation (str, optional): behavior on distance violation. can be "error", "warn", None (silent), or "ignore" (do nothing, keep invalid meshpoints). Defaults to "error".
            radiative_correction (bool, optional): Whether to add raditative correction for cross section calculations. Defaults to False.
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).

        Raises:
            ValueError: No mesh points at y=0, Invalid material config
        """
        super().__init__(device=device, **kwargs)
        self.mesh = "2D"
        self.n_dim = 2

        # test for collisions:
        geo = torch.as_tensor(positions, dtype=DTYPE_FLOAT, device=self.device)

        if on_distance_violation.lower() == "ignore":
            geo_clean = geo
        if step is not None:
            norm = torch.norm(geo.unsqueeze(0) - geo.unsqueeze(1), dim=-1)
            norm[norm.triu() == 0] += 100 * step
            geo_clean = geo[norm.min(dim=0).values >= step * 0.999]
        else:
            warnings.warn("step not provided, cannot check mesh consistency.")
            geo_clean = geo

        if on_distance_violation.lower() == "error" and (len(geo) > len(geo_clean)):
            raise ValueError(
                "{} meshpoints in structure are too close!".format(
                    len(geo) - len(geo_clean)
                )
            )
        elif on_distance_violation.lower() == "warn" and (len(geo) > len(geo_clean)):
            warnings.warn(
                "{} meshpoints in structure are too close! Removing concerned meshpoints and continue.".format(
                    len(geo) - len(geo_clean)
                )
            )
        self.positions = torch.as_tensor(
            geo_clean, dtype=DTYPE_FLOAT, device=self.device
        )

        if torch.count_nonzero(self.positions[..., 1]) > 0:
            warnings.warn("require 2D structure. Remove all positions with y!=0.")
            self.positions = self.positions[self.positions[..., 1] != 0]
            if len(self.positions) == 0:
                raise ValueError("No mesh positions at y=0. Please check geometry.")

        self.r0 = self.get_center_of_mass()  # center of gravity

        if step is None:
            step_scalar = get_step_from_geometry(self.positions)
        else:
            step_scalar = step

        # step for every meshcell, for consistency with other struct classes
        self.step = step_scalar * torch.ones(
            len(self.positions), dtype=DTYPE_FLOAT, device=self.device
        )

        self.radiative_correction = radiative_correction
        self.mesh_normalization_factor = torch.as_tensor(
            1, dtype=DTYPE_FLOAT, device=self.device
        )  # square mesh

        # material of each meshpoint
        if hasattr(materials, "__iter__"):
            if len(materials) != len(self.positions):
                raise ValueError(
                    "Either a global material needs to be given or "
                    + "each meshpoint needs a defined material. "
                    + "But meshpoint list and materials list are of different lengths."
                )
            self.materials = materials
        else:
            self.materials = [materials for i in self.positions]

        self.zeros = torch.zeros(
            (len(self.positions), 3, 3), dtype=DTYPE_COMPLEX, device=self.device
        )

        # discretized, made from natural material: only electric response
        self.interaction_type = "E"  # possible types "E" or "EH"

    def __repr__(self, verbose=False):
        """description about structure"""
        out_str = ""
        out_str += "------ discretized 2D nano-object -------"
        out_str += "\n" + " mesh type:              {}".format(self.mesh)
        out_str += "\n" + " nr. of meshpoints:      {}".format(len(self.positions))
        out_str += "\n" + " nominal stepsizes (nm): {}".format(
            [float(f) for f in torch.unique(self.step)]
        )
        out_str += "\n" + " material:               {}".format(
            [m.__name__ for m in set(self.materials)]
        )
        bnds = ptp(self.positions, dim=0)
        out_str += "\n" + " size & position (Y-axis is infinite):"
        out_str += "\n" + "     X-extension          :   {:.1f} (nm)".format(bnds[0])
        out_str += "\n" + "     Z-extension          :   {:.1f} (nm)".format(bnds[2])
        out_str += "\n" + "     center of mass (x,z) : ({:.1f}, {:.1f})".format(
            self.r0[0], self.r0[2]
        )

        return out_str

    def set_device(self, device):
        """move all tensors of the class to device"""
        super().set_device(device)

        self.zeros = self.zeros.to(device)
        self.step = self.step.to(device)
        self.r0 = self.r0.to(device)
        self.mesh_normalization_factor = self.mesh_normalization_factor.to(device)

        for mat in self.materials:
            mat.set_device(device)

    # --- self-terms
    def get_selfterm_pE(self, wavelength: float, environment) -> torch.Tensor:
        """return list of 'EE' self-term tensors (3x3) at each meshpoint

        Args:
            wavelength (float): in nm
            environment (environment class): 2D environement class

        Returns:
            torch.Tensor: pE self term tensor
        """
        eps_env = environment.get_environment_permittivity_scalar(
            wavelength, r_probe=self.positions
        )
        # cast env permittivity to real, because hankel only support real args
        eps_env = torch.as_tensor(eps_env, dtype=DTYPE_COMPLEX, device=self.device).real

        k0 = 2 * torch.pi / wavelength
        k02 = k0**2

        if self.mesh_normalization_factor == 0:
            norm_xz = 0
            norm_y = 0
        else:
            from torchgdm.tools.special import H1n

            S = self.step**2
            k0_y = environment.get_k0_y(wavelength)

            kr2 = torch.as_tensor(
                eps_env * k02 - k0_y**2, dtype=DTYPE_FLOAT, device=self.device
            )
            kr = torch.sqrt(kr2)

            h11 = H1n(1, kr * self.step / torch.pi**0.5)
            norm01 = self.step / torch.pi**0.5 * h11 / kr + 2j / (torch.pi * kr**2)

            norm_xz_nonrad = -1 * self.mesh_normalization_factor / (2.0 * S * eps_env)
            norm_xz_rad = 1j * torch.pi * (2 * k02 - kr2 / eps_env) * norm01 / (4 * S)

            norm_y_nonrad = 0
            norm_y_rad = 1j * torch.pi * (k02 - k0_y**2 / eps_env) * norm01 / (2 * S)

            norm_xz = 4.0 * torch.pi * (norm_xz_nonrad + norm_xz_rad)
            norm_y = 4.0 * torch.pi * (norm_y_nonrad + norm_y_rad)

        self_terms_pE = torch.zeros(
            (len(self.positions), 3, 3), dtype=DTYPE_COMPLEX, device=self.device
        )
        self_terms_pE[:, 0, 0] = norm_xz
        self_terms_pE[:, 1, 1] = norm_y
        self_terms_pE[:, 2, 2] = norm_xz

        return self_terms_pE

    # --- polarizabilities
    def get_polarizability_pE(self, wavelength: float, environment) -> torch.Tensor:
        """return list of EE polarizability tensors (3x3) at each meshpoint

        Args:
            wavelength (float): in nm
            environment (environment class): 2D environement class

        Returns:
            torch.Tensor: pE polarizability tensor
        """
        eps_env = environment.get_environment_permittivity_scalar(
            wavelength, r_probe=self.positions
        )
        eps_env_tensor = torch.zeros(
            (len(self.positions), 3, 3), dtype=DTYPE_COMPLEX, device=self.device
        )
        eps_env_tensor[:, 0, 0] = eps_env
        eps_env_tensor[:, 1, 1] = eps_env
        eps_env_tensor[:, 2, 2] = eps_env

        eps_geo = torch.zeros(
            (len(self.positions), 3, 3), dtype=DTYPE_COMPLEX, device=self.device
        )
        for i, mat in enumerate(self.materials):
            eps_geo[i] = mat.get_epsilon(wavelength)

        S_cell_norm = self.step**2 / float(self.mesh_normalization_factor)

        # --- polarizability
        alpha_pE = (
            (eps_geo - eps_env_tensor)
            * S_cell_norm.unsqueeze(1).unsqueeze(1)
            / (4.0 * torch.pi)
        )

        return alpha_pE

    # - radiative correction for cross section calc. - 2D case
    def get_radiative_correction_prefactor_p(self, wavelength: float, environment):
        """return electric dipole radiative correction prefactor vectors (3) at each meshpoint"""
        if self.radiative_correction:
            k0 = 2 * torch.pi / wavelength

            pf_vec = torch.as_tensor(
                [1.0, 2.0, 1.0], device=self.device, dtype=DTYPE_COMPLEX
            )
            pf_vec = pf_vec * torch.pi / 2 * k0**2

            return torch.ones(
                (len(self.get_all_positions()), 3),
                device=self.device,
                dtype=DTYPE_COMPLEX,
            ) * pf_vec.unsqueeze(0)
        else:
            return torch.zeros(
                (len(self.get_all_positions()), 3),
                device=self.device,
                dtype=DTYPE_COMPLEX,
            )

    def get_radiative_correction_prefactor_m(self, wavelength: float, environment):
        """return magnetic dipole radiative correction prefactor vectors (3) at each meshpoint"""
        if self.radiative_correction:
            k0 = 2 * torch.pi / wavelength
            n_env = (
                environment.get_environment_permittivity_scalar(
                    wavelength, self.get_all_positions()
                )
                ** 0.5
            )
            pf_vec = torch.as_tensor(
                [1.0, 2.0, 1.0], device=self.device, dtype=DTYPE_COMPLEX
            )
            pf_vec = pf_vec * torch.pi / 2 * k0**2

            return (
                torch.ones(
                    (len(self.get_all_positions()), 3),
                    device=self.device,
                    dtype=DTYPE_COMPLEX,
                )
                * pf_vec.unsqueeze(0)
                * n_env.unsqueeze(1) ** 2
            )
        else:
            return torch.zeros(
                (len(self.get_all_positions()), 3),
                device=self.device,
                dtype=DTYPE_COMPLEX,
            )

    # --- plotting
    def plot(self, **kwargs):
        """plot the structure in XZ plane (2D)"""
        from torchgdm.visu import visu2d

        kwargs["projection"] = "xz"
        im = visu2d.geo2d._plot_structure_discretized(self, **kwargs)
        return im

    def plot_contour(self, **kwargs):
        """plot the structure contour in XZ plane (2D)"""
        from torchgdm.visu import visu2d

        kwargs["projection"] = "xz"
        im = visu2d.geo2d._plot_contour_discretized(self, **kwargs)
        return im

    def plot3d(self, **kwargs):
        """plot the structure in 3D"""
        from torchgdm.visu import visu3d

        return visu3d.geo3d._plot_structure_discretized(self, **kwargs)

    # --- geometry operations
    def rotate(
        self,
        alpha: float,
        center: torch.Tensor = torch.as_tensor([0.0, 0.0, 0.0]),
        axis: str = "y",
    ):
        """rotate the structure

        Args:
            alpha (float): rotation angle (in rad)
            center (torch.Tensor, optional): center of rotation axis. Defaults to torch.as_tensor([0.0, 0.0, 0.0]).
            axis (str, optional): rotation axis. Defaults to "y".

        Raises:
            ValueError: only "y" axis supported in 2D

        Returns:
            :class:`StructDiscretized2D`: copy of structure with rotated geometry
        """
        _struct_rotated = self.copy()
        center = center.to(dtype=DTYPE_FLOAT, device=self.device)

        if axis.lower() == "y":
            rot = rotation_y(alpha, device=self.device)
        else:
            raise ValueError(
                "Only rotation axis 'y' supported in 2D (infinite axis).".format(axis)
            )

        if len(_struct_rotated.positions) > 1:
            _struct_rotated.positions = torch.matmul(
                _struct_rotated.positions - (center + self.r0), rot
            ) + (center + self.r0)
        else:
            warnings.warn("Single meshpoint found, ignore rotation.")

        return _struct_rotated

    # --- effective model extraction wrapper
    def convert_to_effective_polarizability_pair(
        self, wavelengths, environment=None, test_accuracy=False, **kwargs
    ):
        from torchgdm.struct.struct2d import StructEffPola2D

        # from torchgdm.struct.struct2d.line import extract_eff_pola_2d
        from torchgdm.struct.eff_model_tools import extract_eff_pola_2d
        from torchgdm.struct.eff_model_tools import _test_effective_model_accuracy

        warnings.warn(
            "2D effective polarizabilities only implemented for illumination incidence in XZ plane!"
        )
        wavelengths = torch.as_tensor(
            wavelengths, dtype=DTYPE_FLOAT, device=self.device
        )
        wavelengths = torch.atleast_1d(wavelengths)

        alpha = extract_eff_pola_2d(
            struct=self, wavelengths=wavelengths, environment=environment, **kwargs
        )
        struct_aeff = StructEffPola2D(
            positions=torch.stack([self.r0]),
            alpha_dicts=[alpha],
            device=self.device,
        )

        if test_accuracy:
            _test_results = _test_effective_model_accuracy(struct_aeff, self)

        return struct_aeff

    def convert_to_gpm(
        self,
        wavelengths,
        r_gpm,
        skeletonize=False,
        r_probe=None,
        illumination_list=None,
        environment=None,
        probe_type="particle",
        n_probe=1500,
        probe_spacing=5.0,
        n_planewave_sources=5,
        dipole_source_type="particle",
        n_dipole_sources=60,
        dipole_sources_spacing=6.0,
        verbose=True,
        progress_bar=True,
        device=None,
        residual_warning_threshold=0.1,
        test_accuracy=False,
        **kwargs,
    ):
        """convert the structure to a global polarizability matrix (GPM)

            Extract the GPM in a given `environement` at `wavelengths`.
            This is done in 3 steps:

                1) Illuminate with various sources, calculated scattered fields at
                various probe positions
                2) The effective dipole moment for each GPM dipole is obtained
                via matching of their emission and the probe fields of (1).
                3) A second inverse problem of adjusting the GPM to create the dipoles
                moments found in (2) is solved via pseudoinverse.

            By default, use a mix of plane waves and local dipole sources (different incidence directions, polarizations, locations, orientations).

        Args:
            struct (class:`StructDiscretized3D`): the discretized structure to extract the model for.
            wavelengths (torch.Tensor): list of wavelengths to extract the model at. in nm.
            r_gpm (int or torch.Tensor): Either int: number of positions or a torch.Tensor of shape (N, 3), giving the locations where to place effective dipole pairs.
            skeletonize (bool, optional): Has effect only if `r_gpm` is of type int. If True, perform a skeletonization prior clustering. Defaults to False.
            r_probe (torch.Tensor): probe positions where the simulated and the GPM scattered fields are matched. Overrides all other probe-related configurations. If not given, use automatic probe positions, which may not be optimal. Defaults to None (automatic).
            illumination_list (list of illuminations): List of torchgdm illumination fields to use for extraction. If not given, use automatic illumination fields, which may not be optimal. Defaults to None (automatic).
            environment (environment class): 3D environement class. Defaults to None (vacuum).
            probe_type (str, optional): Where to probe, one of ["particle", "circle"]. "particle": probe at fixed distance to particle surface. "circle": probe on enclosing circle surface. Defaults to "particle".
            n_probe (int, optional): maximum number of probe positions on enclosing circle. Defaults to 1500.
            probe_spacing (float, optional): additional distance to particle or to enclsing circle surface, in units of discretization step. Defaults to 5.0.
            n_planewave_sources (int, optional): number of plane wave angles to use as illumination. Defaults to 7.
            dipole_source_type (str, optional): Where to put sources, one of ["particle", "circle"]. "particle": probe at fixed distance to particle surface. "circle": sources on enclosing circle. Defaults to "particle".
            n_dipole_sources (int, optional): If given, use `n` dipole sources as illumination instead of plane waves. Defaults to None.
            dipole_sources_spacing (float, optional): if using dipole light sources, additional distance to enclosing circle surface, in units of discretization step. Defaults to  6.0.
            verbose (bool, optional): whether to sprint progess info. Defaults to True.
            progress_bar (bool, optional): whether to show a progress bar. Defaults to True.
            device (str, optional). Defaults to None, in which case the structure's device is used.
            residual_warning_threshold (float, optional). Maximum residual L2 norm before triggering a warning. Defaults to 0.1.

            Returns:
                :class:`torchgdm.struct.struct3d.gpm.StructGPM3D`: Effective dipole-pair model
        """
        from torchgdm.struct.struct2d.gpm2d import StructGPM2D
        from torchgdm.struct.eff_model_tools import extract_gpm_from_struct
        from torchgdm.struct.eff_model_tools import _test_effective_model_accuracy

        wavelengths = torch.as_tensor(
            wavelengths, dtype=DTYPE_FLOAT, device=self.device
        )
        wavelengths = torch.atleast_1d(wavelengths)

        gpm_dict = extract_gpm_from_struct(
            struct=self,
            wavelengths=wavelengths,
            r_gpm=r_gpm,
            skeletonize=skeletonize,
            r_probe=r_probe,
            illumination_list=illumination_list,
            environment=environment,
            probe_type=probe_type,
            n_probe=n_probe,
            probe_spacing=probe_spacing,
            n_planewave_sources=n_planewave_sources,
            dipole_source_type=dipole_source_type,
            n_dipole_sources=n_dipole_sources,
            dipole_sources_spacing=dipole_sources_spacing,
            verbose=verbose,
            progress_bar=progress_bar,
            device=device,
            residual_warning_threshold=residual_warning_threshold,
            **kwargs,
        )

        struct_gpm = StructGPM2D(
            positions=self.r0,
            gpm_dicts=gpm_dict,
            device=self.device,
            progress_bar=progress_bar,
        )

        # perform a test in an actual scattering simulation
        if test_accuracy:
            _test_results = _test_effective_model_accuracy(
                struct_gpm, self, which=["ecs", "nf_sca"]
            )

        return struct_gpm


class StructDiscretizedSquare2D(StructDiscretized2D):
    """class for square surface discretized, infinitely long 2D structure

    Defines the square surface discretization, polarizabilities and self-terms
    """

    __name__ = "2D square lattice discretized structure class"

    def __init__(
        self,
        discretization_config,
        step,
        materials,
        device: torch.device = None,
        **kwargs,
    ):
        """2D structure, discretized on a square lattice

        Infinite axis along Y

        Args:
            discretization_config (tuple): tuple of discretization condition function, and discretizer walk limits (as provided by the geometries generators)
            step (float, optional): nominal stepsize. If not provided, infer automatically from geometry. Defaults to None.
            materials (material class, or list of them): material of structure. If list, define the material of each meshpoint.
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).

        Raises:
            ValueError: No mesh points at y=0, Invalid material config
        """
        from torchgdm.struct.struct2d import discretizer_square

        positions = discretizer_square(*discretization_config, step=step)

        super().__init__(
            positions,
            materials,
            step=step,
            device=device,
            on_distance_violation="ignore",  # trust the discretizer --> speedup
            **kwargs,
        )

        self.mesh = "square (2D)"
        self.mesh_normalization_factor = torch.tensor(
            1.0, dtype=DTYPE_FLOAT, device=self.device
        )



# -*- coding: utf-8 -*-
"""
handling of linear equation systems (coupling)
"""
# %%
from collections.abc import Callable
import time
import warnings

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import _purge_mem, get_default_device
from torchgdm.tools.batch import batched


def _reduce_dimensions(Q: torch.Tensor):
    """expand dimensions of Q from (A1, A2, N, N) to (A1*N, A2*N)

    Args:
        Q (torch.Tensor): 4D tensor of shape (A1, A2, N, N), containing an array of (N, N) tensors

    Returns:
        torch.Tensor: tensor of shape (A1*N, A2*N) with conserved global structure of Q
    """
    Q_red = Q.permute(0, 2, 1, 3)
    dim_Q1 = Q_red.shape[0] * Q_red.shape[1]
    dim_Q2 = Q_red.shape[2] * Q_red.shape[3]

    # note: this is not memory efficient as it copies the full matrix...
    Q_red = Q_red.contiguous().view(dim_Q1, dim_Q2)
    return Q_red


def _expand_dimensions(Q: torch.Tensor, N: int = 6):
    """expand dimensions of Q from (A1*N, A2*N) to (A1, A2, N, N)

    Args:
        Q (torch.Tensor): 4D tensor of shape (A1, A2, N, N), containing an array of (N, N) tensors
        N (int): size of rank 2 sub-tensors. Defaults to 6

    Returns:
        torch.Tensor: tensor of shape (A1, A2, N, N) with conserved global structure of Q
    """
    dim_Q1 = Q.shape[0] // N
    dim_Q2 = Q.shape[1] // N

    # note: this is not memory efficient as it copies the full matrix...
    Q_exp = Q.contiguous().view(dim_Q1, N, dim_Q2, N)
    Q_exp = Q_exp.permute(0, 2, 1, 3)

    return Q_exp


@batched("f0")
def _batched_lu_solve(LU, pivots, f0=None):
    if f0 is None:
        raise ValueError("'f0' must be given.")
    return torch.linalg.lu_solve(LU, pivots, f0)


# --- linear system base class
class LinearSystemBase:
    """torchgdm base class for linearsystem, describing and solving the coupling"""

    __name__ = "linearsystem base class"

    def __init__(self, device: torch.device = None):
        if device is None:
            self.device = get_default_device()
        else:
            self.device = device

    def set_device(self, device):
        """move all tensors of the class to device"""
        self.device = device

    def solve(self):
        """solve coupling for illumination(s)"""
        raise NotImplementedError("`solve` not implemented with this linear system.")

    def get_generalized_propagator(self):
        """get the generalized propagator"""
        raise NotImplementedError(
            "Generalized propoagator not available with this linear system."
        )


class LinearSystemFullInverse(LinearSystemBase):
    """Solve coupling by full inversion

    This linear system solver class is a reference implementation using full 6x6 coupling at
    every position, even if not all polarizability tensor components are required.
    In cases containing full volume discretization structures, this is not memory and compute efficient .
    """

    def __init__(self, device: torch.device = None):
        """Solve coupling by full inversion

        Args:
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).
        """
        super().__init__(device=device)

    def set_device(self, device):
        """move all tensors of the class to device"""
        super().set_device(device)

    # --- functions to setup interaction systems
    def _get_full_interaction_matrix_G_tensors(
        self, positions: torch.Tensor, G_func: Callable, wavelength: float
    ) -> torch.Tensor:
        # evaluate Green's tensors for each couple of points
        # iteract_G_NxN is of shape (N,N,6,6)
        interact_G_NxN = G_func(
            positions.unsqueeze(1),
            positions.unsqueeze(0),
            wavelength=wavelength,
        )
        return interact_G_NxN

    def _get_full_Gdotalpha(
        self,
        sim,
        G_func: Callable,
        wavelength: float,
    ) -> torch.Tensor:
        """calc. interaction matrix $(G.alpha)$, selfterms on diagonal"""

        positions = sim.get_all_positions()
        all_gpm = []
        for s in sim.structures:
            all_gpm += list(
                s.get_polarizability_6x6(wavelength, sim.environment).unbind()
            )
        block_gpm = torch.block_diag(*all_gpm)

        all_selfterms = []
        for s in sim.structures:
            all_selfterms += list(
                s.get_selfterm_6x6(wavelength, sim.environment).unbind()
            )
        block_selfterms = torch.block_diag(*all_selfterms)

        ones_gpm = torch.block_diag(
            *[torch.ones_like(s.real).to(torch.int) for s in all_gpm]
        )

        N = len(positions)

        interact_NxNx6x6 = self._get_full_interaction_matrix_G_tensors(
            positions, G_func, wavelength
        )
        interact_NxN = _reduce_dimensions(interact_NxNx6x6)
        interact_NxN = interact_NxN.masked_fill(ones_gpm == 1, 0)
        interact_NxN += block_selfterms

        G_dot_interact = torch.matmul(interact_NxN, block_gpm)

        return G_dot_interact

    def get_interact(self, sim_gpm, wavelength: float) -> torch.Tensor:
        """get the full interaction block (E, H) of the linear system for a list of structures

        containes all direct and cross-contributions (E-E, H-H, E-H, H-E)
        """
        # calc. interactions (G dot GPM)
        interact = self._get_full_Gdotalpha(
            sim=sim_gpm,
            G_func=sim_gpm.environment.get_G_6x6,
            wavelength=wavelength,
        )

        # invertible matrix block is (1 - G.alpha), which adds the illumination field
        interact = torch.eye(len(interact), device=self.device) - interact

        return interact

    def solve(
        self,
        sim,
        wavelength: float,
        batch_size=32,
        verbose: int = 1,
    ) -> torch.Tensor:
        """solve system of coupled structures for all illuminations

        Args:
            sim (SimulationBase): simulation container
            wavelength (float): wavelength (nm)
            batch_size (int, optional): Nr of incident fields evaluated in parallel. Defaults to 32.
            verbose (int, optional): wether to print status info. Defaults to 1.

        Returns:
            torch.Tensor: list of internal fields, one for each illumination. Shape: (N_e0, nr of pos, 3)
        """
        if verbose >= 2:
            t_start = time.time()
            print("linsys", end="")

        # - setup interaction system
        interact = self.get_interact(sim, wavelength)
        if verbose >= 2:
            t_interact = time.time()
            print(" {:.2f}s. solve".format(t_interact - t_start), end="")

        # - inversion
        # LU, pivots, info = torch.linalg.lu_factor_ex(interact)  # experimental version, faster on GPU
        LU, pivots = torch.linalg.lu_factor(interact)
        if verbose >= 2:
            t_lu = time.time()
            print(
                " {:.2f}s. {}xE0".format(
                    t_lu - t_interact, len(sim.illumination_fields)
                ),
                end="",
            )

        # - evaluate zero order fields
        f0 = sim._get_all_e0_h0(wavelength)
        f0 = f0.view(len(f0), -1, 1)

        # - solve for RHS vectors
        f_inside = _batched_lu_solve(LU, pivots, f0=f0, batch_size=batch_size)
        eh_inside = f_inside.view(len(f_inside), -1, 6)  # conv 6N to N 6-tuples (E,H)
        e_inside, h_inside = torch.chunk(eh_inside, 2, dim=2)  # separate E and H fields

        if verbose >= 2:
            t_solve = time.time()
            print(" {:.2f}s.".format(t_solve - t_lu), end="")

        return e_inside, h_inside

    def get_generalized_propagator(self, sim, wavelength: float) -> torch.Tensor:
        interact = self.get_interact(sim, wavelength)
        K = torch.linalg.inv(interact)
        return K


class LinearSystemFullMemEff(LinearSystemFullInverse):
    """Solve coupling by full inversion

    This linear system solver couples only actually polarizable terms.
    """

    def __init__(self, device: torch.device = None):
        """Solve coupling by full inversion

        This linear system solver couples only actually polarizable terms.

        Args:
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).
        """
        super().__init__(device=device)

    def set_device(self, device):
        """move all tensors of the class to device"""
        super().set_device(device)

    # --- select proper coupling for structure pair
    def get_struct_pair_coupling_config(
        self,
        struct1,
        struct2,
        sim,
        verbose=False,
    ):
        """get evaluation functions for required coupling between two structures

        Concerns Greens tensor, polarizability and self-term evaluation

        Args:
            struct1 (structBase): target of scattered light
            struct2 (structBase): origin of light scattering
            sim (SimulationBase): simulation descriptor

        Returns:
            callable: Green's tensor calculation function
            callable: polarizability calculation function
            callable: self-term evaluator, only returned for equal coupling of st1 and st2. default: None
        """

        _s1T = struct1.interaction_type  # observer structure (target)
        _s2T = struct2.interaction_type  # emitter structure (origin)
        selfterm_func = (
            None  # default: two different structures --> no self-terms required
        )

        # --- electric-only observer and source
        # e <- e
        if _s1T == "E" and _s2T == "E":
            G_func = sim.environment.get_G_Ep
            alpha_func = struct2.get_polarizability_pE
            selfterm_func = struct2.get_selfterm_pE

        # --- mixed observer, electric-only source
        # e,m <- e
        elif _s1T == "EH" and _s2T == "E":
            G_func = sim.environment.get_G_6x6
            alpha_func = struct2.get_polarizability_pmE_6x3

        # --- electric-only observer, mixed source
        # e <- e,m
        elif _s1T == "E" and _s2T == "EH":
            G_func = sim.environment.get_G_Epm_3x6
            alpha_func = struct2.get_polarizability_6x6

        # --- full coupling case
        # e,m <- e,m
        else:
            G_func = sim.environment.get_G_6x6
            alpha_func = struct2.get_polarizability_6x6
            selfterm_func = struct2.get_selfterm_6x6

        return G_func, alpha_func, selfterm_func

    # --- functions to setup interaction systems
    def _get_block_Gdotalpha(
        self,
        struct1: torch.Tensor,
        struct2: torch.Tensor,
        sim,
        wavelength: float,
    ) -> torch.Tensor:
        """calc. invertible interaction matrix $(G.alpha)$. selfterms on diagonal

        consider light scattering from source (struct2) to target (struct1)
        """
        # process selfterms and polarizabilities
        G_func, alpha_func, selfterm_func = self.get_struct_pair_coupling_config(
            struct1, struct2, sim
        )

        # all polarizabilities
        all_pola = list(alpha_func(wavelength, sim.environment).unbind())
        block_pola = torch.block_diag(*all_pola)
        ones_gpm = torch.block_diag(
            *[torch.ones_like(s.real).to(torch.int) for s in all_pola]
        )

        # Green's tensors form s2 to s1. shape (N1,N2,6,6)
        interact_N1xN2x6x6 = G_func(
            struct1.get_all_positions().unsqueeze(1),  # light targets
            struct2.get_all_positions().unsqueeze(0),  # light sources
            wavelength=wavelength,
        )
        interact_N1xN2 = _reduce_dimensions(interact_N1xN2x6x6)

        # self-terms
        if struct1 == struct2:
            all_selfterms = list(selfterm_func(wavelength, sim.environment).unbind())
            block_selfterms = torch.block_diag(*all_selfterms)

            # fill GPMs (nonlocal polas, coupling already solved)
            interact_N1xN2 = interact_N1xN2.masked_fill(ones_gpm == 1, 0)
            interact_N1xN2 += block_selfterms

        G_dot_interact = torch.matmul(interact_N1xN2, block_pola)

        return G_dot_interact

    def get_interact(self, sim, wavelength: float, verbose=False) -> torch.Tensor:
        """get the full interaction block (E, H) of the linear system for a list of structures

        containes all direct and cross-contributions (E-E, H-H, E-H, H-E)

        Note: This implementation is not efficient for large numbers of structures.
        TODO: combine all structures of same interaction type before doing the simulation
        """
        from torchgdm.tools.misc import tqdm

        prgbar = True if verbose >= 3 else False

        # combine consecutive structures of same coupling type to reduce
        # memory transfer cost during iteration of all structures
        struct_list_reduced = [sim.structures[0]]
        if len(sim.structures) > 1:
            for _s in tqdm(
                sim.structures[1:], progress_bar=prgbar, title="optimizing structures"
            ):
                if _s.interaction_type == struct_list_reduced[-1].interaction_type:
                    try:
                        struct_list_reduced[-1] = struct_list_reduced[-1].combine(_s)
                    except:
                        # if combine fails, add as separate structure
                        struct_list_reduced.append(_s)
                else:
                    # if different coupling type, add as separate structure
                    struct_list_reduced.append(_s)

        if len(struct_list_reduced) > 10:
            warnings.warn(
                "Structure list could not be fully optimized. "
                + "For optimal performance, add structures with same "
                + "mesh or same coupling type in consecutive groups."
            )

        # calc. G dot alpha
        interact_rows = []
        for struct2 in tqdm(
            struct_list_reduced, progress_bar=prgbar, title="interaction matrix"
        ):
            interact_cols = []
            for struct1 in struct_list_reduced:
                # interaction block for struct2 (emitter) --> struct1 (observer)
                interact_expand = self._get_block_Gdotalpha(
                    struct1=struct1,
                    struct2=struct2,
                    sim=sim,
                    wavelength=wavelength,
                )
                interact_cols.append(interact_expand)
            interact_rows.append(torch.cat(interact_cols, dim=0))
        interact_full = torch.cat(interact_rows, dim=1)

        # free (GPU-)RAM
        _purge_mem(interact_rows)

        # invertible matrix block is (1 - G.alpha): add the illumination field to output
        interact_full = (
            torch.eye(len(interact_full), device=self.device) - interact_full
        )

        return interact_full

    # --- field mask operations
    def _zero_fill_nonpolarizable_fields(
        self, sim, field_at_polarizable: torch.Tensor
    ) -> torch.Tensor:
        """create full field tensor with polarizable fields inserted.

        Args:
            sim (`torchgdm.Simulation`): simulation instance
            field_at_polarizable (torch.Tensor): field values to be inserted at polarizable positions

        Returns:
            torch.Tensor: full field tensor at all meshpoints with masked regions inserted
        """
        # init full internal e/h field tensor (N_inc, N_pos, 6)
        fields_full = torch.zeros(
            (len(sim.illumination_fields), len(sim.get_all_positions()), 6),
            dtype=DTYPE_COMPLEX,
            device=self.device,
        )

        # get polarizable positions mask for field tensor
        mask_fields = sim._get_polarizable_mask_full_fields()

        # insert field elements at polarizable positions into masked full field tensor
        fields_full.flatten()[mask_fields.flatten()] = field_at_polarizable.flatten()

        return fields_full

    def solve(
        self,
        sim,
        wavelength: float,
        batch_size=32,
        verbose: int = 1,
    ) -> torch.Tensor:
        """solve system of coupled structures for all illuminations

        Args:
            sim (`torchgdm.Simulation`): simulation instance
            wavelength (float): wavelength (nm)
            batch_size (int, optional): Nr of incident fields evaluated in parallel. Defaults to 32.
            verbose (int, optional): verbose level of status reporting. One of [0,1,2]. The higher, the more info are printed. Defaults to 1.

        Returns:
            torch.Tensor: list of internal fields, one for each illumination. Shape: (N_e0, nr of pos, 3)
        """
        if verbose >= 2:
            t_start = time.time()
            print("linsys", end="")

        # - setup interaction system
        interact = self.get_interact(sim, wavelength, verbose=verbose)
        if verbose >= 2:
            t_interact = time.time()
            print(" {:.2f}s. solve".format(t_interact - t_start), end="")

        # - inversion
        # LU, pivots, info = torch.linalg.lu_factor_ex(interact)  # experimental version, faster on GPU
        LU, pivots = torch.linalg.lu_factor(interact)
        if verbose >= 2:
            t_lu = time.time()
            print(
                " {:.2f}s. {}xE0".format(
                    t_lu - t_interact, len(sim.illumination_fields)
                ),
                end="",
            )

        # - evaluate zero order fields (illumination)
        f0 = sim._get_polarizablefields_e0_h0(wavelength)
        f0 = f0.view(len(f0), -1, 1)  # add dim for lu_solve

        # - solve for RHS vectors
        f_masked = _batched_lu_solve(LU, pivots, f0=f0, batch_size=batch_size)
        eh_inside = self._zero_fill_nonpolarizable_fields(sim, f_masked[..., 0])
        e_inside, h_inside = torch.chunk(eh_inside, 2, dim=2)  # separate E and H fields

        if verbose >= 2:
            t_solve = time.time()
            print(" {:.2f}s.".format(t_solve - t_lu), end="")

        return e_inside, h_inside

    def get_generalized_propagator(self, sim, wavelength: float) -> torch.Tensor:
        """solve system of coupled structures without illumination

        Args:
            sim (`torchgdm.Simulation`): simulation instance
            wavelength (float): wavelength (nm)

        Returns:
            torch.Tensor: full generalized propagator tensor for all polarizable elements
        """
        interact = self.get_interact(sim, wavelength)
        K = torch.linalg.inv(interact)
        return K

## --- legacy

class _LinearSystemFullInverse(LinearSystemBase):
    """Solve coupling by full inversion

    This linear system solver class is a non-optimized reference implementation for testing.
    It uses (6x6) polarizabilities for all dipoles, which is non memory efficient in most cases.
    """

    def __init__(self, device: torch.device = None):
        """Solve coupling by full inversion

        Memory-inefficient test implementation. Uses 6x6 polarizabilities for every positio.

        Args:
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).
        """
        super().__init__(device=device)

    def set_device(self, device):
        """move all tensors of the class to device"""
        super().set_device(device)

    # --- functions to setup interaction systems
    def _get_full_interaction_matrix_G_tensors(
        self, positions: torch.Tensor, G_func: Callable, wavelength: float
    ) -> torch.Tensor:
        # evaluate Green's tensors for each couple of points
        # iteract_G_NxN is of shape (N,N,3,3)
        interact_G_NxN = G_func(
            positions.unsqueeze(1),
            positions.unsqueeze(0),
            wavelength=wavelength,
        )
        return interact_G_NxN

    def _get_full_Gdotalpha(
        self,
        positions: torch.Tensor,
        polarizabilities: torch.Tensor,
        self_terms: torch.Tensor,
        G_func: Callable,
        wavelength: float,
    ) -> torch.Tensor:
        """calc. invertible interaction matrix $(G.alpha)$, selfterms on diagonal"""
        N = len(positions)

        interact_NxN = self._get_full_interaction_matrix_G_tensors(
            positions, G_func, wavelength
        )

        # put self-term matrices on diagonal (at R1[i]==R2[i])
        eye_indices = torch.arange(N, device=self.device)
        interact_NxN[eye_indices, eye_indices, :, :] = self_terms.unsqueeze(0)

        # multiply by polarizabilities
        interact_NxN = torch.matmul(interact_NxN, polarizabilities.unsqueeze(0))

        # reshape (N,N,6,6) to (6N, 6N) matrix (or (3N,3N) if electric-only)
        interact_6Nx6N = _reduce_dimensions(interact_NxN)

        return interact_6Nx6N

    def get_interact(self, sim, wavelength: float) -> torch.Tensor:
        """get the full interaction block (E, H) of the linear system for a list of structures

        containes all direct and cross-contributions (E-E, H-H, E-H, H-E)
        """
        # calc. interactions (G.alpha)
        all_positions = sim.get_all_positions()

        interact = self._get_full_Gdotalpha(
            positions=all_positions,
            polarizabilities=sim._get_all_polarizabilitites_6x6(wavelength),
            self_terms=sim._get_all_selfterms_6x6(wavelength),
            G_func=sim.environment.get_G_6x6,
            wavelength=wavelength,
        )

        # invertible matrix block is (1 - G.alpha), which adds the illumination field
        interact = torch.eye(len(interact), device=self.device) - interact

        return interact

    def solve(
        self,
        sim,
        wavelength: float,
        batch_size=32,
        verbose: int = 1,
    ) -> torch.Tensor:
        """solve system of coupled structures for all illuminations

        Args:
            sim (SimulationBase): simulation container
            wavelength (float): wavelength (nm)
            batch_size (int, optional): Nr of incident fields evaluated in parallel. Defaults to 32.
            verbose (int, optional): wether to print status info. Defaults to 1.

        Returns:
            torch.Tensor: list of internal fields, one for each illumination. Shape: (N_e0, nr of pos, 3)
        """
        if verbose >= 2:
            t_start = time.time()
            print("linsys", end="")

        # - setup interaction system
        interact = self.get_interact(sim, wavelength)
        if verbose >= 2:
            t_interact = time.time()
            print(" {:.2f}s. solve".format(t_interact - t_start), end="")

        # - inversion
        # LU, pivots, info = torch.linalg.lu_factor_ex(interact)  # experimental version, faster on GPU
        LU, pivots = torch.linalg.lu_factor(interact)
        if verbose >= 2:
            t_lu = time.time()
            print(
                " {:.2f}s. {}xE0".format(
                    t_lu - t_interact, len(sim.illumination_fields)
                ),
                end="",
            )

        # - evaluate zero order fields
        f0 = sim._get_all_e0_h0(wavelength)
        f0 = f0.view(len(f0), -1, 1)

        # - solve for RHS vectors
        f_inside = _batched_lu_solve(LU, pivots, f0=f0, batch_size=batch_size)
        eh_inside = f_inside.view(len(f_inside), -1, 6)  # conv 6N to N 6-tuples (E,H)
        e_inside, h_inside = torch.chunk(eh_inside, 2, dim=2)  # separate E and H fields

        if verbose >= 2:
            t_solve = time.time()
            print(" {:.2f}s.".format(t_solve - t_lu), end="")

        return e_inside, h_inside

    def get_generalized_propagator(self, sim, wavelength: float) -> torch.Tensor:
        interact = self.get_interact(sim, wavelength)
        K = torch.linalg.inv(interact)
        return K


class _LinearSystemFullMemEffnoGPM(LinearSystemFullInverse):
    """Solve coupling by full inversion

    This linear system solver couples only actually polarizable terms.
    """

    def __init__(self, device: torch.device = None):
        """Solve coupling by full inversion

        This linear system solver couples only actually polarizable terms.

        Args:
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).
        """
        super().__init__(device=device)

    def set_device(self, device):
        """move all tensors of the class to device"""
        super().set_device(device)

    # --- select proper coupling for structure pair
    def get_struct_pair_coupling_config(
        self,
        struct1,
        struct2,
        sim,
        verbose=False,
    ):
        """get evaluation functions for required coupling between two structures

        Concerns Greens tensor, polarizability and self-term evaluation

        Args:
            struct1 (structBase): target of scattered light
            struct2 (structBase): origin of light scattering
            sim (SimulationBase): simulation descriptor

        Returns:
            callable: Green's tensor calculation function
            callable: polarizability calculation function
            callable: self-term evaluator, only returned for equal coupling of st1 and st2. default: None
        """

        _s1T = struct1.interaction_type  # observer structure (target)
        _s2T = struct2.interaction_type  # emitter structure (origin)
        selfterm_func = (
            None  # default: two different structures --> no self-terms required
        )

        # --- electric-only observer and source
        # e <- e
        if _s1T == "E" and _s2T == "E":
            G_func = sim.environment.get_G_Ep
            alpha_func = struct2.get_polarizability_pE
            selfterm_func = struct2.get_selfterm_pE

        # --- mixed observer, electric-only source
        # e,m <- e
        elif _s1T == "EH" and _s2T == "E":
            G_func = sim.environment.get_G_6x6
            alpha_func = struct2.get_polarizability_pmE_6x3

        # --- electric-only observer, mixed source
        # e <- e,m
        elif _s1T == "E" and _s2T == "EH":
            G_func = sim.environment.get_G_Epm_3x6
            alpha_func = struct2.get_polarizability_6x6

        # --- full coupling case
        # e,m <- e,m
        else:
            G_func = sim.environment.get_G_6x6
            alpha_func = struct2.get_polarizability_6x6
            selfterm_func = struct2.get_selfterm_6x6

        return G_func, alpha_func, selfterm_func

    # --- functions to setup interaction systems
    def _get_block_Gdotalpha(
        self,
        struct1: torch.Tensor,
        struct2: torch.Tensor,
        sim,
        wavelength: float,
    ) -> torch.Tensor:
        """calc. invertible interaction matrix $(G.alpha)$. selfterms on diagonal"""

        # process selfterms and polarizabilities
        G_func, alpha_func, selfterm_func = self.get_struct_pair_coupling_config(
            struct1, struct2, sim
        )

        # eval. Greens tensors between struct2 (emitter) and struct1 (observer)
        gdotalpha = G_func(
            struct1.get_all_positions().unsqueeze(1),  # light targets
            struct2.get_all_positions().unsqueeze(0),  # light sources
            wavelength=wavelength,
        )

        # if struct1 == struct2: put self-terms on diagonal (at R1[i]==R2[i])
        if struct1 == struct2:
            self_terms = selfterm_func(wavelength, sim.environment)

            N = len(struct1.get_all_positions())
            eye_idx = torch.arange(N, device=self.device)
            gdotalpha[eye_idx, eye_idx, :, :] = self_terms.unsqueeze(0)

        # multiply by polarizabilities
        polarizabilities = alpha_func(wavelength, sim.environment)
        gdotalpha = torch.matmul(gdotalpha, polarizabilities.unsqueeze(0))

        # reshape (N1,N2,6,6) to (6xN1, 6xN2) matrix (or (3,3), (6,3), (3,6) if not full coupling)
        gdotalpha_expanded = _reduce_dimensions(gdotalpha)

        return gdotalpha_expanded

    def get_interact(self, sim, wavelength: float, verbose=False) -> torch.Tensor:
        """get the full interaction block (E, H) of the linear system for a list of structures

        containes all direct and cross-contributions (E-E, H-H, E-H, H-E)

        Note: This implementation is not efficient for large numbers of structures.
        TODO: combine all structures of same interaction type before doing the simulation
        """
        from torchgdm.tools.misc import tqdm

        prgbar = True if verbose >= 3 else False

        # combine consecutive structures of same coupling type to reduce
        # memory transfer cost during iteration of all structures
        struct_list_reduced = [sim.structures[0]]
        if len(sim.structures) > 1:
            for _s in tqdm(
                sim.structures[1:], progress_bar=prgbar, title="optimizing structures"
            ):
                if _s.interaction_type == struct_list_reduced[-1].interaction_type:
                    try:
                        struct_list_reduced[-1] = struct_list_reduced[-1].combine(_s)
                    except:
                        # if combine fails, add as separate structure
                        struct_list_reduced.append(_s)
                else:
                    # if different coupling type, add as separate structure
                    struct_list_reduced.append(_s)

        if len(struct_list_reduced) > 10:
            warnings.warn(
                "Structure list could not be fully optimized. "
                + "For optimal performance, add structures with same "
                + "mesh or same coupling type in consecutive groups."
            )

        # calc. G dot alpha
        interact_rows = []
        for struct2 in tqdm(
            struct_list_reduced, progress_bar=prgbar, title="interaction matrix"
        ):
            interact_cols = []
            for struct1 in struct_list_reduced:
                # interaction block for struct2 (emitter) --> struct1 (observer)
                interact_expand = self._get_block_Gdotalpha(
                    struct1=struct1,
                    struct2=struct2,
                    sim=sim,
                    wavelength=wavelength,
                )
                interact_cols.append(interact_expand)
            interact_rows.append(torch.cat(interact_cols, dim=0))
        interact_full = torch.cat(interact_rows, dim=1)

        # free (GPU-)RAM
        _purge_mem(interact_rows)

        # invertible matrix block is (1 - G.alpha): add the illumination field to output
        interact_full = (
            torch.eye(len(interact_full), device=self.device) - interact_full
        )

        return interact_full

    # --- field mask operations
    def _zero_fill_nonpolarizable_fields(
        self, sim, field_at_polarizable: torch.Tensor
    ) -> torch.Tensor:
        """create full field tensor with polarizable fields inserted.

        Args:
            sim (`torchgdm.Simulation`): simulation instance
            field_at_polarizable (torch.Tensor): field values to be inserted at polarizable positions

        Returns:
            torch.Tensor: full field tensor at all meshpoints with masked regions inserted
        """
        # init full internal e/h field tensor (N_inc, N_pos, 6)
        fields_full = torch.zeros(
            (len(sim.illumination_fields), len(sim.get_all_positions()), 6),
            dtype=DTYPE_COMPLEX,
            device=self.device,
        )

        # get polarizable positions mask for field tensor
        mask_fields = sim._get_polarizable_mask_full_fields()

        # insert field elements at polarizable positions into masked full field tensor
        fields_full.flatten()[mask_fields.flatten()] = field_at_polarizable.flatten()

        return fields_full

    def solve(
        self,
        sim,
        wavelength: float,
        batch_size=32,
        verbose: int = 1,
    ) -> torch.Tensor:
        """solve system of coupled structures for all illuminations

        Args:
            sim (`torchgdm.Simulation`): simulation instance
            wavelength (float): wavelength (nm)
            batch_size (int, optional): Nr of incident fields evaluated in parallel. Defaults to 32.
            verbose (int, optional): verbose level of status reporting. One of [0,1,2]. The higher, the more info are printed. Defaults to 1.

        Returns:
            torch.Tensor: list of internal fields, one for each illumination. Shape: (N_e0, nr of pos, 3)
        """
        if verbose >= 2:
            t_start = time.time()
            print("linsys", end="")

        # - setup interaction system
        interact = self.get_interact(sim, wavelength, verbose=verbose)
        if verbose >= 2:
            t_interact = time.time()
            print(" {:.2f}s. solve".format(t_interact - t_start), end="")

        # - inversion
        # LU, pivots, info = torch.linalg.lu_factor_ex(interact)  # experimental version, faster on GPU
        LU, pivots = torch.linalg.lu_factor(interact)
        if verbose >= 2:
            t_lu = time.time()
            print(
                " {:.2f}s. {}xE0".format(
                    t_lu - t_interact, len(sim.illumination_fields)
                ),
                end="",
            )

        # - evaluate zero order fields (illumination)
        f0 = sim._get_polarizablefields_e0_h0(wavelength)
        f0 = f0.view(len(f0), -1, 1)  # add dim for lu_solve

        # - solve for RHS vectors
        f_masked = _batched_lu_solve(LU, pivots, f0=f0, batch_size=batch_size)
        eh_inside = self._zero_fill_nonpolarizable_fields(sim, f_masked[..., 0])
        e_inside, h_inside = torch.chunk(eh_inside, 2, dim=2)  # separate E and H fields

        if verbose >= 2:
            t_solve = time.time()
            print(" {:.2f}s.".format(t_solve - t_lu), end="")

        return e_inside, h_inside

    def get_generalized_propagator(self, sim, wavelength: float) -> torch.Tensor:
        """solve system of coupled structures without illumination

        Args:
            sim (`torchgdm.Simulation`): simulation instance
            wavelength (float): wavelength (nm)

        Returns:
            torch.Tensor: full generalized propagator tensor for all polarizable elements
        """
        interact = self.get_interact(sim, wavelength)
        K = torch.linalg.inv(interact)
        return K



# -*- coding: utf-8 -*-
"""
core field container class
"""
# %%
import warnings

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import tqdm


# Container for fields
class Field:
    """container for electro-magnetic field data

    contains:
     - evaluation points
     - if applicable: line elements or surface elements

    routines:
     - field intensity
     - intensity integration
     - chirality
     - Poynting
     - plotting

    """

    def __init__(
        self,
        positions: torch.Tensor,
        efield: torch.Tensor,
        hfield: torch.Tensor = None,
        ds: torch.Tensor = None,
        device=None,
    ):
        """container for electro-magnetic field data

        Args:
            positions (torch.Tensor): positions at which field data exists. Either a tensor of shape (N,3), or a dict containing "r_probe" and "ds" keys.
            efield (torch.Tensor): complex electric field at positions. Shape must match the positions.
            hfield (torch.Tensor, optional): complex magnetic field at positions. Shape must match the positions. If not given, set to zero (e.g. for far-fields). Defaults to None.
            ds (torch.Tensor, optional): optional surface (or line) elements corresponding to each position. If given, used in field integration. Defaults to None.
            device (str, optional): If None, use device of `efield` tensor. Defaults to None.

        Raises:
            ValueError: insufficient inputs provided
        """
        # exception handling
        if efield is None and hfield is None:
            raise ValueError("At least one of efield or hfield needs to be given.")

        # set device
        if device is None:
            self.device = efield.device
        else:
            self.device = device

        # main data
        self.efield = efield
        self.hfield = hfield
        if type(positions) == dict:
            self.ds = positions["ds"] if (ds is None and "ds" in positions) else ds
            self.positions = positions["r_probe"]
        else:
            self.ds = ds
            self.positions = positions

        # set not given data to zero / unit
        if self.ds is None:  # surface elements (for integration)
            self.ds = torch.ones(
                len(self.positions), device=self.device, dtype=DTYPE_FLOAT
            )
        if self.efield is None:
            self.efield = torch.zeros_like(self.hfield)
        if self.hfield is None:
            self.hfield = torch.zeros_like(self.efield)

        # move tensors to same device as fields
        self.set_device(self.device)

        assert self.efield.shape[1:] == self.positions.shape
        assert self.hfield.shape[1:] == self.positions.shape
        assert len(self.ds) == len(self.positions)

    def __add__(self, other):
        if issubclass(type(other), Field):
            # if all positions are identical, superpose the fields
            if len(self.positions) == len(other.positions):
                if torch.all(self.positions == other.positions):
                    new_field = self.copy()
                    new_field.set_device(other.device)
                    new_field.efield = new_field.efield + other.efield
                    new_field.hfield = new_field.hfield + other.hfield

                    return new_field
            # if no points are shared, combine all positions
            if (
                len(
                    (
                        (self.positions[:, None, :] == other.positions[None, ...]).all(
                            dim=2
                        )
                    ).nonzero()
                )
                == 0
            ):
                all_pos = torch.cat([self.positions, other.positions], dim=0)
                all_e = torch.cat([self.efield, other.efield], dim=1)
                all_h = torch.cat([self.hfield, other.hfield], dim=1)
                all_ds = torch.cat([self.ds, other.ds], dim=0)
                new_field = Field(
                    positions=all_pos,
                    efield=all_e,
                    hfield=all_h,
                    ds=all_ds,
                    device=other.device,
                )
                return new_field

            raise ValueError(
                "Addition only possible if either "
                + "all positions or none of the positions are equal."
            )
        else:
            raise ValueError("Addition only possible with `Field` instances.")

    def __sub__(self, other):
        if issubclass(type(other), Field):
            assert torch.all(self.positions == other.positions)

            new_field = self.copy()
            new_field.set_device(other.device)
            new_field.efield = new_field.efield - other.efield
            new_field.hfield = new_field.hfield - other.hfield

            return new_field
        else:
            raise ValueError("Subtraction only possible with `Field` instances.")

    def __repr__(self, verbose: bool = False):
        """description about field object"""
        from torchgdm.tools.misc import ptp

        bnds = ptp(self.positions, dim=0)
        out_str = "------ E/M field container -------"
        out_str += "\n" + " nr of fields: {}".format(len(self.efield)) + "\n"
        out_str += " grid ({} positions):".format(len(self.positions)) + "\n"
        out_str += "   Nx = {:}, ".format(len(torch.unique(self.positions[:, 0])))
        out_str += "Ny = {:}, ".format(len(torch.unique(self.positions[:, 1])))
        out_str += "Nz = {:}, ".format(len(torch.unique(self.positions[:, 2])))

        out_str += "\n" + " extension (nm) : " + "\n"
        out_str += "   X  = {:.1f}, ".format(bnds[0])
        out_str += "Y  = {:.1f}, ".format(bnds[1])
        out_str += "Z  = {:.1f}, ".format(bnds[2])

        return out_str

    def cat(self, other):
        """concatenate with other field

        will add another complex field "layer" of same size
        (Nfields1, Nx, Ny), (Nfields2, Nx, Ny) --> (Nfields1 + Nfields2, Nx, Ny)
        """
        if issubclass(type(other), Field) and torch.all(
            self.positions == other.positions
        ):
            new_field = self.copy()
            new_field.set_device(other.device)
            new_field.efield = torch.cat([new_field.efield, other.efield], dim=0)
            new_field.hfield = torch.cat([new_field.hfield, other.hfield], dim=0)

            return new_field

        else:
            raise ValueError("Concatenation only possible all positions are equal.")

    def copy(self):
        import copy

        return copy.deepcopy(self)

    def set_device(self, device):
        self.positions = self.positions.to(device=device)
        self.efield = self.efield.to(device=device)
        self.ds = self.ds.to(device=device)
        if self.hfield is not None:
            self.hfield = self.hfield.to(device=device)

    def set_efield(self, efield: torch.Tensor):
        """replace the electric field by `efield`"""
        assert efield.shape[1:] == self.positions.shape
        self.efield = torch.as_tensor(efield, device=self.device)

    def set_hfield(self, hfield: torch.Tensor):
        """replace the magnetic field by `hfield`"""
        assert hfield.shape[1:] == self.positions.shape
        self.hfield = torch.as_tensor(hfield, device=self.device)

    def set_ds(self, ds: torch.Tensor):
        """replace the surface elements by `ds`"""
        assert len(ds) == len(self.positions)
        self.ds = torch.as_tensor(ds, device=self.device)

    # --- postprocessing methods
    def get_positions(self, **kwargs):
        """return all positions"""

        return self.positions

    def get_efield(self, warn=True, **kwargs):
        """return the electric field at all positions"""
        if warn and torch.sum(torch.abs(self.efield)) == 0:
            warnings.warn("No e-fields available.")

        return self.efield

    def get_hfield(self, warn=True, **kwargs):
        """return the magnetic field at all positions"""
        if warn and torch.sum(torch.abs(self.hfield)) == 0:
            warnings.warn("No h-fields available.")

        return self.hfield

    def get_efield_intensity(self, warn=True, **kwargs):
        """return the electric field intensity at all positions"""
        if warn and torch.sum(torch.abs(self.efield)) == 0:
            warnings.warn("No e-fields available.")

        I_e = torch.sum(torch.abs(self.efield) ** 2, dim=-1)
        return I_e

    def get_hfield_intensity(self, warn=True, **kwargs):
        """return the magnetic field intensity at all positions"""
        if warn and torch.sum(torch.abs(self.hfield)) == 0:
            warnings.warn("No h-fields available.")

        I_h = torch.sum(torch.abs(self.hfield) ** 2, dim=-1)
        return I_h

    def get_poynting(self, **kwargs):
        """get the complex Poynting vector at all positions

        S = conj(E) x H
        (conj(): complex conjugate)
        """
        if torch.sum(torch.abs(self.efield)) == 0:
            warnings.warn("No e-fields available.")
        if torch.sum(torch.abs(self.hfield)) == 0:
            warnings.warn("No h-fields available.")

        S = torch.cross(torch.conj(self.efield), self.hfield, dim=-1)
        return S

    def get_energy_flux(self, **kwargs):
        """time average Poynting vector at all positions

        <S>=Re(S)
        """
        return self.get_poynting().real

    def get_chirality(self, **kwargs):
        """return the near-field chirality at all positions

        calculated as in Meinzer et al, PRB 88, 041407, 2013:
        C ~ Im(conj(E) * B)
        """
        if torch.sum(torch.abs(self.efield)) == 0:
            warnings.warn("No e-fields available.")
        if torch.sum(torch.abs(self.hfield)) == 0:
            warnings.warn("No h-fields available.")

        C = (
            -1
            * torch.sum(
                torch.multiply(torch.conj(self.efield), self.hfield), dim=-1
            ).imag
        )
        return C

    # --- integration methods
    def get_integrated_efield_intensity(self, **kwargs):
        """get the integrated electric field intensity

        return sum(abs(E)^2 x dS) as a single float value
        """
        if torch.mean(torch.abs(self.efield)) == 1:
            warnings.warn("Surface elements are set to 1.")
        return torch.sum(self.get_efield_intensity() * self.ds, dim=-1)

    def get_integrated_hfield_intensity(self, **kwargs):
        """get the integrated magnetic field intensity

        return sum(abs(H)^2 x dS) as a single float value
        """
        if torch.mean(torch.abs(self.efield)) == 1:
            warnings.warn("Surface elements are set to 1.")
        return torch.sum(self.get_hfield_intensity() * self.ds, dim=-1)

    # --- plotting methods
    # - plotting for e-field
    def plot_efield_amplitude(
        self, illumination_index=0, field_component="x", complex_part="re", **kwargs
    ):
        """plot E-field amplitude using :func:`~torchgdm.visu.visu2d.field_amplitude`"""
        from torchgdm.visu.visu2d import field_amplitude as visu_field_amplitude_2d

        return visu_field_amplitude_2d(
            self,
            illumination_index=illumination_index,
            whichfield="e",
            field_component=field_component,
            complex_part=complex_part,
            **kwargs,
        )

    def plot_efield_vectors(self, illumination_index=0, **kwargs):
        """plot E-field vectors using :func:`~torchgdm.visu.visu2d.vectorfield`"""
        from torchgdm.visu.visu2d import vectorfield as visu_vectorfield_2d

        return visu_vectorfield_2d(
            self, illumination_index=illumination_index, whichfield="e", **kwargs
        )

    def plot_efield_intensity(self, illumination_index=0, **kwargs):
        """plot E-field intensity map using :func:`~torchgdm.visu.visu2d.field_intensity`"""
        from torchgdm.visu.visu2d import field_intensity as visu_field_intensity_2d

        return visu_field_intensity_2d(
            self, illumination_index=illumination_index, whichfield="e", **kwargs
        )

    def plot_efield_vectors3d(self, illumination_index=0, **kwargs):
        """plot E-field 3D vectors using :func:`~torchgdm.visu.visu3d.vectorfield`"""
        from torchgdm.visu.visu3d import vectorfield as visu_vectorfield_3d

        return visu_vectorfield_3d(
            self, illumination_index=illumination_index, whichfield="e", **kwargs
        )

    # - plotting for h-field
    def plot_hfield_amplitude(
        self, illumination_index=0, field_component="x", complex_part="re", **kwargs
    ):
        """plot H-field amplitude using :func:`~torchgdm.visu.visu2d.field_amplitude`"""
        from torchgdm.visu.visu2d import field_amplitude as visu_field_amplitude_2d

        return visu_field_amplitude_2d(
            self,
            illumination_index=illumination_index,
            whichfield="h",
            field_component=field_component,
            complex_part=complex_part,
            **kwargs,
        )

    def plot_hfield_vectors(self, illumination_index=0, **kwargs):
        """plot H-field vectors using :func:`~torchgdm.visu.visu2d.vectorfield`"""
        from torchgdm.visu.visu2d import vectorfield as visu_vectorfield_2d

        return visu_vectorfield_2d(
            self, illumination_index=illumination_index, whichfield="h", **kwargs
        )

    def plot_hfield_intensity(self, illumination_index=0, **kwargs):
        """plot H-field intensity map using :func:`~torchgdm.visu.visu2d.field_intensity`"""
        from torchgdm.visu.visu2d import field_intensity as visu_field_intensity_2d

        return visu_field_intensity_2d(
            self, illumination_index=illumination_index, whichfield="h", **kwargs
        )

    def plot_hfield_vectors3d(self, illumination_index=0, **kwargs):
        """plot H-field 3D vectors using :func:`~torchgdm.visu.visu3d.vectorfield`"""
        from torchgdm.visu.visu3d import vectorfield as visu_vectorfield_3d

        return visu_vectorfield_3d(
            self, illumination_index=illumination_index, whichfield="h", **kwargs
        )

    # - plot chirality
    def plot_chirality(self, illumination_index=0, **kwargs):
        """plot nearfield chirality :func:`~torchgdm.visu.visu2d.`"""
        from torchgdm.visu.visu2d.scalar2d import _scalarfield

        return _scalarfield(
            self.get_chirality()[illumination_index], self.positions, **kwargs
        )

    # - plotting for time averaged Poynting
    def plot_energy_flux_vectors(self, illumination_index=0, **kwargs):
        """plot energy flux vector field using :func:`torchgdm.visu.visu2d.vec2d._vectorfield`"""
        from torchgdm.visu.visu2d.vec2d import _vectorfield

        f = self.get_energy_flux()[illumination_index]

        return _vectorfield(f, self.positions, **kwargs)

    def plot_energy_flux_streamlines(self, illumination_index=0, **kwargs):
        """plot energy flux streamlines using :func:`~torchgdm.visu.visu2d.vec2d.streamlines_energy_flux`"""
        from torchgdm.visu.visu2d.vec2d import streamlines_energy_flux

        return streamlines_energy_flux(
            self, illumination_index=illumination_index, **kwargs
        )



# -*- coding: utf-8 -*-
"""point polarizability classes"""
import warnings
import copy

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device
from torchgdm.struct.base_classes import StructBase
from torchgdm.struct.struct3d.point import StructEffPola3D
from torchgdm.struct.struct2d.surface import StructDiscretized2D
from torchgdm.tools import interp
from torchgdm.tools.geometry import sample_random_circular
from torchgdm.tools.misc import ptp


# --- base class volume discretized structure container - 3D
class StructEffPola2D(StructEffPola3D):
    """class for 2D line polarizability structure (infinite y axis)"""

    __name__ = "effective line polarizability (2D) structure class"

    def __init__(
        self,
        positions: torch.Tensor,
        alpha_dicts: list,
        radiative_correction: bool = True,
        device: torch.device = None,
        environment=None,
        shift_z_to_r0: bool = True,
    ):
        """2D line polarizability class

        The main information is provided in the `alpha_dicts`, which is a list of dicts with the full effective polarizability definitions. Each dict defines one structure and must contain following:
            - 'wavelengths': wavelengths at which the polarizabilities are calculated
            - at least one of: ['alpha_pE', 'alpha_mH', 'alpha_mE', 'alpha_pH']:
                polarizability tensors of shape [len(wavelengths), 3, 3]
            optional keys:
            - 'full_geometry': the original volume discretization of the represented geometry
            - 'r0': the origin of the effective polarizabilities with respect to optional 'full_geometry'
            - 'enclosing_radius': enclosing radius of original structure

        Args:
            positions (torch.Tensor): polarizability positions (3D, but all y values must be zero)
            alpha_dicts (list): list of polarizability model dictionaries
            radiative_correction (bool, optional): Whether to add raditative correction for cross section calculations. Defaults to True.
            device (torch.device, optional): Defaults to "cpu".
            environment (`environment` instance, optional): 2D environment class. By default use environment defined in eff.pola. dictionary. Defaults to None.
            shift_z_to_r0 (bool, optional): If True, if a position z-value is zero, each polarizability model's z position will be shifted to the height of the effective dipole development center. Defaults to True.

        Raises:
            ValueError: positions for 2D simulations must be at y=0
        """
        if device is None:
            self.device = get_default_device()
        else:
            self.device = device

        # expand positions, put single scatterer in list
        _positions = torch.as_tensor(positions, dtype=DTYPE_FLOAT, device=self.device)
        if len(_positions.shape) == 1:
            assert len(_positions) == 3
            _positions = _positions.unsqueeze(0)

        if torch.count_nonzero(_positions[..., 1]) > 0:
            warnings.warn("2D structure. Remove all positions with y!=0.")
            _positions = _positions[_positions[..., 1] != 0]
            if len(_positions) == 0:
                raise ValueError("No mesh positions at y=0. Please check geometry.")

        super().__init__(
            positions=positions,
            alpha_dicts=alpha_dicts,
            environment=environment,
            shift_z_to_r0=shift_z_to_r0,
            device=device,
        )

        self.n_dim = 2

        self.radiative_correction = radiative_correction

    def __repr__(self, verbose=False):
        """description about structure"""
        out_str = ""
        out_str += (
            "------ 2D effective ED / MD line-dipole polarizabilities object -------"
        )
        out_str += "\n" + " nr. of dipole-pairs:    {}".format(
            len(self.get_all_positions())
        )
        out_str += "\n" + " nominal enclosing circle diameters (nm): {}".format(
            [round(float(f), 1) for f in torch.unique(self.step)]
        )
        if len(self.full_geometries) > 0:
            pos = torch.cat(self.full_geometries)
            out_str += "\n" + " original 2D geometry: "
            out_str += "\n" + "  - replacing nr. of meshpoints: {}".format(len(pos))
            bnds = ptp(pos, dim=0)
            out_str += "\n" + "  - size & position:"
            out_str += "\n" + "        X-extension    :    {:.1f} (nm)".format(bnds[0])
            out_str += "\n" + "        Z-extension    :    {:.1f} (nm)".format(bnds[2])
            out_str += "\n" + "  - center of mass :    ({:.1f}, {:.1f}, {:.1f})".format(
                *[float(f) for f in self.get_center_of_mass()]
            )

        return out_str

    # - radiative correction for cross section calc. - 2D case
    # inherit from discretized 2D class
    get_radiative_correction_prefactor_p = (
        StructDiscretized2D.get_radiative_correction_prefactor_p
    )
    get_radiative_correction_prefactor_m = (
        StructDiscretized2D.get_radiative_correction_prefactor_m
    )

    # --- plotting
    def plot(
        self,
        scale=1.0,
        color="auto",
        linestyle_circle=(0, (2, 2)),
        color_circle="auto",
        color_circle_fill=None,
        alpha=1,
        show_grid=True,
        color_grid="auto",
        alpha_grid=0.25,
        legend=True,
        set_ax_aspect=True,
        reset_color_cycle=True,
        **kwargs,
    ):
        """plot the structure of the effective line-polarizability (2D)

        Args:
            scale (float, optional): scaling factor of the grid cells, if shown. Defaults to 1.0.
            color (str, optional): plot color. Defaults to "auto".
            linestyle_circle (tuple, optional): optional line style for enclosing circle. Defaults to (0, (2, 2)).
            color_circle (str, optional): optional alternative color for enclosing circle. Defaults to "auto".
            color_circle_fill (_type_, optional): optional alternative fill color for enclosing circle. Defaults to None.
            alpha (int, optional): optional transparency. Defaults to 1.
            show_grid (bool, optional): whether to show mesh grid (if available in structure). Defaults to True.
            color_grid (str, optional): optional alternative color for the mesh grid. Defaults to "auto".
            alpha_grid (float, optional): optional alternative transparency for the mesh grid. Defaults to 0.25.
            legend (bool, optional): show legend. Defaults to True.
            set_ax_aspect (bool, optional): automatically set aspect ratio to equal. Defaults to True.
            reset_color_cycle (bool, optional): reset color cycle after finishing the plot. Defaults to True.

        Returns:
            matplotlib axes
        """
        from torchgdm.visu import visu2d

        kwargs["projection"] = "xz"
        im = visu2d.geo2d._plot_structure_eff_pola(
            self,
            scale=scale,
            color=color,
            linestyle_circle=linestyle_circle,
            color_circle=color_circle,
            color_circle_fill=color_circle_fill,
            alpha=alpha,
            show_grid=show_grid,
            color_grid=color_grid,
            alpha_grid=alpha_grid,
            legend=legend,
            set_ax_aspect=set_ax_aspect,
            reset_color_cycle=reset_color_cycle,
            **kwargs,
        )
        return im

    def plot_contour(
        self,
        color="auto",
        set_ax_aspect=True,
        alpha=1.0,
        alpha_value=None,
        reset_color_cycle=True,
        **kwargs,
    ):
        """plot the contour of the underlying 2D-mesh (2D)

        Args:
            color (str, optional): optional matplotlib compatible color. Defaults to "auto".
            set_ax_aspect (bool, optional): If True, will set aspect of plot to equal. Defaults to True.
            alpha (float, optional): matplotlib transparency value. Defaults to 1.0.
            alpha_value (float, optional): alphashape value. If `None`, try to automatically optimize for best enclosing of structure. Defaults to None.
            reset_color_cycle (bool, optional): reset color cycle after finishing the plot. Defaults to True.

        Returns:
            matplotlib line: matplotlib's `scatter` output
        """
        from torchgdm.visu import visu2d

        kwargs["projection"] = "xz"
        im = visu2d.geo2d.contour(
            self,
            color=color,
            set_ax_aspect=set_ax_aspect,
            alpha=alpha,
            alpha_value=alpha_value,
            reset_color_cycle=reset_color_cycle,
            **kwargs,
        )
        return im

    def plot3d(self, **kwargs):
        """plot the point polarizability structure (3D)"""
        from torchgdm.visu import visu3d

        warnings.warn(
            "Visualizing a 2D model in 3D will show only meshpoints in the 'XZ' plane. The 3D plot shows a circumscribing sphere but is in fact a circumscribing cylinder."
        )

        return visu3d.geo3d._plot_structure_eff_3dpola(self, **kwargs)

    # --- geometry operations
    def rotate(
        self,
        alpha: float,
        center: torch.Tensor = torch.as_tensor([0.0, 0.0, 0.0]),
        axis: str = "y",
    ):
        """rotate the structure

        Args:
            alpha (float): rotation angle (in rad)
            center (torch.Tensor, optional): center of rotation axis. Defaults to torch.as_tensor([0.0, 0.0, 0.0]).
            axis (str, optional): rotation axis. Defaults to "y".

        Raises:
            ValueError: only "y" axis supported in 2D

        Returns:
            :class:`StructDiscretized2D`: copy of structure with rotated geometry
        """
        if axis.lower() != "y":
            raise ValueError(
                "Only rotation axis 'y' supported in 2D (infinite axis).".format(axis)
            )

        _struct_rotated = super().rotate(alpha=alpha, center=center, axis=axis)

        return _struct_rotated


# --- Mie sphere
class StructMieCylinderEffPola2D(StructEffPola2D):
    """class for Mie-theory based 2D line polarizability

    Requires external package `treams`
    !!! class constructor does not support automatic differentiation !!!

    Defines a 2D line polarizability representing a cylinder using
    first order (dipolar) Mie coefficients
    """

    __name__ = "Mie-theory cylinder line-dipole polarizability (2D) structure class"

    def __init__(
        self,
        wavelengths: torch.Tensor,
        radii: list,
        materials: list,
        environment=None,
        r0: torch.Tensor = None,
        device: torch.device = None,
        quadrupol_tol=0.15,
    ):
        """2D line polarizability class for a core-shell cylinder (Mie)

        Use Mie theory in dipole approximation (first order) to get an
        effective polarizability model for a core-shell cylinder.
        Requires the `treams` package for Mie coefficient calculation.
        https://github.com/tfp-photonics/treams

        Args:
            wavelengths (torch.Tensor): list of wavelengths to evaluate (nm)
            radii (list): list of the cylinders's core and (multiple) shell radii (in nm).
            materials (list): materials of core and shell(s). A float or int is interpreted as permittivity value.
            environment (environment instance, optional): Homogeneous 3D environment to evaluate Mie theory in. Defaults to None, which uses vacuum.
            r0 (torch.Tensor, optional): polarizability position (x,y,z). If not given, is set to (0, 0, 0). Defaults to None
            device (torch.device, optional): Defaults to "cpu".
            quadrupol_tol (float, optional): ratio of tolerable residual quadrupole terms relative to the dipole order before warning. Defaults to 0.15.

        Raises:
            ValueError: incorrect parameters
        """
        # prep and imports
        from torchgdm.tools.misc import to_np
        from torchgdm.materials.base_classes import MaterialBase
        from torchgdm.materials import MatConstant
        import numpy as np

        # TODO: Replace with differentiable Mie code
        from torchgdm.tools.mie import mie_ab_cylinder_2d

        if device is None:
            self.device = get_default_device()
        else:
            self.device = device

        # tensor conversion
        wavelengths = torch.as_tensor(
            wavelengths, dtype=DTYPE_FLOAT, device=self.device
        )
        wavelengths = torch.atleast_1d(wavelengths)
        k0 = 2 * torch.pi / wavelengths

        # mie coefficients
        mie_results = mie_ab_cylinder_2d(
            wavelengths=wavelengths,
            radii=radii,
            materials=materials,
            environment=environment,
            device=self.device,
            m_max=3,
            as_dict=True,
        )
        a_n = mie_results["a_n"]
        b_n = mie_results["b_n"]
        env = mie_results["environment"]
        n_env = mie_results["n_env"]
        r_enclosing = mie_results["r_enclosing"]

        # check if dipole approximation is good
        a_quadrupol_res = a_n[:, 2].abs()
        wls_violation_a = to_np(wavelengths[a_quadrupol_res.to("cpu") > quadrupol_tol])
        if len(wls_violation_a) > 0:
            warnings.warn(
                "Mie series: {} wavelengths with ".format(len(wls_violation_a))
                + "significant residual electric quadrupole contribution: "
                + "{} nm".format([round(r, 1) for r in wls_violation_a])
            )

        b_quadrupol_res = b_n[:, 2].abs()
        wls_violation_b = to_np(wavelengths[b_quadrupol_res.to("cpu") > quadrupol_tol])
        if len(wls_violation_b) > 0:
            warnings.warn(
                "Mie series: {} wavelengths with ".format(len(wls_violation_b))
                + "significant residual magnetic quadrupole contribution: "
                + "{} nm".format([round(r, 1) for r in wls_violation_b])
            )

        # convert to polarizabilities (units of volume)
        # populate 6x6 polarizabilities for all wavelengths
        alpha_6x6 = torch.zeros(
            (len(wavelengths), 6, 6), dtype=DTYPE_COMPLEX, device=self.device
        )
        # TM polarization (s-pol., E parallel axis)
        a_par_pE = 1j / torch.pi * a_n[:, 0] / k0**2
        a_par_mH = 1j * 2 / torch.pi * b_n[:, 0] / k0**2 / n_env**2
        alpha_6x6[:, 1, 1] = a_par_pE
        alpha_6x6[:, 3, 3] = a_par_mH
        alpha_6x6[:, 5, 5] = a_par_mH

        # TE polarization (p-pol., E perp. axis)
        a_perp_pE = 1j * 2 / torch.pi * b_n[:, 1] / k0**2
        a_perp_mH = 1j / torch.pi * a_n[:, 1] / k0**2 / n_env**2
        alpha_6x6[:, 0, 0] = a_perp_pE
        alpha_6x6[:, 2, 2] = a_perp_pE
        alpha_6x6[:, 4, 4] = a_perp_mH

        # set center of mass
        if r0 is None:
            r0 = torch.as_tensor([0.0, 0.0, 0.0], dtype=DTYPE_FLOAT, device=self.device)
        else:
            r0 = torch.as_tensor(r0, dtype=DTYPE_FLOAT, device=self.device)
            r0 = r0.squeeze()
            assert len(r0) == 3
            assert r0[1] == 0  # 2D: must be in XZ plane

        # wrap up in a dictionary compatible with the point dipole structure class
        alpha_dict = dict(
            r0=r0,
            r0_MD=r0,
            r0_ED=r0,
            alpha_6x6=alpha_6x6,
            wavelengths=wavelengths,
            enclosing_radius=r_enclosing,
            k0_spectrum=k0,
            environment=env,
        )

        # - point polarizability structure with Mie dipolar response
        super().__init__(
            positions=r0,
            alpha_dicts=[alpha_dict],
            device=self.device,
        )




def _extract_eff_pola_2d_legacy(
    struct,
    wavelengths,
    environment=None,
    n_probe=150,
    distance_probe=10.0,
    n_dipoles=None,
    distance_dipoles=5000,
    verbose=True,
    only_pE_mH=False,
    progress_bar=True,
    device=None,
    batch_size=256,
):
    """Extract effective electric and magnetic dipole polarizability for volume discretized structure

    Extract the polarizability for the structure `struct` in a given `environement`
    at the specified `wavelengths`

    In this version, the effective dipole response for several illuminations is obtained
    via matching on a circumscribing sphere by solving of a first inverse problem via pseudoinverse.

    The second inverse problem of adjusting polarizability for different
    illuminations is also solved via pseudoinverse.

    By default, use 14 plane waves (different incidence directions and polarizations).
    alternative: illumination with `n_dipoles` point-dipole sources if `n_dipoles` is an integer > 0.


    Args:
        struct (class:`StructDiscretized3D`): the discretized structure to extract the model for.
        wavelengths (torch.Tensor): list of wavelengths to extract the model at. in nm.
        environment (environment class, optional): 2D environement class. Defaults to None (vacuum).
        n_probe (int, optional): number of probe positions on enclosing sphere. Defaults to 100.
        distance_probe (float, optional): additional distance to enclosing sphere in units of discretization step. Defaults to 2.0.
        n_dipoles (int, optional): If given, use `n` dipole sources as illumination instead of plane waves. Defaults to None.
        distance_dipoles (int, optional): if using dipoles, specify their distance to the center of gravity. Defaults to 5000.
        verbose (bool, optional): whether to sprint progess info. Defaults to True.
        only_pE_mH (bool, optional): whether to extract only a p/m model (True) or a full (6x6) polarizability (False). Defaults to True.
        progress_bar (bool, optional): whether to show a progress bar. Defaults to True.
        device (str, optional). Defaults to None, in which case the structure's device is used.

    Returns:
        dict: contains all informations about the effective dipole polarizability model
    """
    from torchgdm.env.freespace_2d.inc_fields import ElectricLineDipole, PlaneWave
    from torchgdm.simulation import Simulation
    from torchgdm.postproc.fields import nf
    from torchgdm.tools.geometry import get_enclosing_sphere_radius
    from torchgdm.tools.misc import tqdm
    from torchgdm.tools.misc import _check_environment
    from torchgdm.struct.eff_model_tools import _test_residual_effective_polarizability

    _struct = struct.copy()

    if device is None:
        device = _struct.device
    else:
        struct.set_device(device)

    # convert single wavelength to list
    wavelengths = torch.as_tensor(wavelengths, dtype=DTYPE_FLOAT, device=device)
    wavelengths = torch.atleast_1d(wavelengths)

    # environment
    env = _check_environment(environment, N_dim=2, device=device)

    if verbose:
        import time

        print(
            "--- extracting eff. dipole pair model (2D, {} wavelengths) ---".format(
                len(wavelengths)
            )
        )
        t_start = time.time()

    # use first order multipole moments (propto local field)
    step_max = torch.max(struct.step)
    enclosing_radius = get_enclosing_sphere_radius(_struct.get_all_positions())
    illumination_radius = enclosing_radius + distance_dipoles
    probe_radius = enclosing_radius + distance_probe * step_max
    r0 = struct.get_center_of_mass()

    # setup perpendicular plane waves illuminations
    if n_dipoles is None:
        angles = torch.linspace(-torch.pi, torch.pi * 0.9, 14)
        pw_conf_list = [[1.0, 0.0, angle, "xz"] for angle in angles]
        pw_conf_list += [[0.0, 1.0, angle, "xz"] for angle in angles]
        e_inc_extract = [
            PlaneWave(e0s=a, e0p=b, inc_angle=c, inc_plane=d, device=device)
            for [a, b, c, d] in pw_conf_list
        ]
    # optional: multiple dipole illuminations
    else:
        if n_dipoles <= 0 or type(n_dipoles) != int:
            raise ValueError(
                "dipole illumination mode: `n_dipoles` needs to be a positive integer."
            )

        # setup dipoles of random position and random orientation
        rnd_pos_dp = (
            sample_random_circular(n_dipoles, device=device) * illumination_radius
        )
        e_inc_extract = [
            ElectricLineDipole(
                r_source=r_dp,
                p_source=torch.rand(3, device=device) * illumination_radius,
                device=device,
            )
            for r_dp in rnd_pos_dp
        ]

    # setup field probe positions on enclosing sphere
    r_probe = sample_random_circular(n_probe, device=device) * probe_radius

    # replace illumination
    _sim = Simulation(
        structures=[_struct],
        environment=env,
        illumination_fields=e_inc_extract,
        wavelengths=wavelengths,
        device=device,
    )

    if verbose:
        t0 = time.time()
        _pos_p, _pos_m = _sim._get_polarizable_positions_p_m()
        n_dp = len(_pos_p) + len(_pos_m)
        n_wl = len(wavelengths)
        print("Running simulation ({} dipoles, {} wls)... ".format(n_dp, n_wl), end="")
    _sim.run(verbose=False, progress_bar=progress_bar)

    if verbose and not progress_bar:
        print("Done in {:.2f}s.".format(time.time() - t0))

    # solve the optimization problem
    alpha_6x6 = torch.zeros(
        (len(wavelengths), 6, 6), dtype=DTYPE_COMPLEX, device=device
    )
    if verbose:
        t0 = time.time()
        print("Running p/m optimization... ", end="")

    for i_wl in tqdm(range(len(wavelengths)), progress_bar, title=""):
        wl = wavelengths[i_wl]

        # calculate fields on enclosing sphere for all illuminations
        # shape: (n_illumination, n_probe*6)
        nf_probe = nf(
            _sim, wl, r_probe=r_probe, progress_bar=False, batch_size=batch_size
        )
        e_eval = nf_probe["sca"].get_efield()
        h_eval = nf_probe["sca"].get_hfield()
        f_eval = torch.cat([e_eval, h_eval], dim=2)
        f_eval = f_eval.reshape(len(e_inc_extract), -1)

        # illuminating fields at expansion location
        # shape: (n_illumination, 6)
        e0_eval = torch.zeros(
            (len(e_inc_extract), 3), dtype=DTYPE_COMPLEX, device=device
        )
        h0_eval = torch.zeros(
            (len(e_inc_extract), 3), dtype=DTYPE_COMPLEX, device=device
        )
        for i_field, e_inc in enumerate(e_inc_extract):
            inc_f = e_inc.get_field(r0.unsqueeze(0), wl, env)
            e0_eval[i_field] = inc_f.get_efield()
            h0_eval[i_field] = inc_f.get_hfield()

        f0_eval = torch.cat([e0_eval, h0_eval], dim=1)

        # --- full 6x6 polarizability
        if not only_pE_mH:
            # calculate Green's tensors between r0 and r_probe
            # shape: (n_probe*6, 6)
            G_6x6 = env.get_G_6x6(r_probe=r_probe, r_source=r0, wavelength=wl)
            G_all = G_6x6.reshape(-1, 6)

            # inv. problem #1: probe fields + Green's tensors --> dipole moments
            Gi = torch.linalg.pinv(G_all)
            pm_eff = torch.matmul(Gi.unsqueeze(0), f_eval.unsqueeze(-1))[..., 0]

            # inv. problem #2: dipole moments + illumination --> effective pola
            pinv_f0 = torch.linalg.pinv(f0_eval)
            alpha_6x6_inv = torch.matmul(pinv_f0, pm_eff)

            # optimum alphas to obtain dipole moments for each illumination
            alpha_6x6[i_wl] = alpha_6x6_inv

            # test residuals
            _test_residual_effective_polarizability(
                _struct,
                wl,
                env,
                alpha_6x6_inv,
                f0_eval,
                pm_eff,
                text_which_dp="6x6 dipole",
            )

        # --- only pE and mH
        if only_pE_mH:
            # calculate Green's tensors between r0 and r_probe
            # shape: (n_probe*6, 6)
            G_Ep = env.get_G_Ep(r_probe=r_probe, r_source=r0, wavelength=wl)
            G_Hm = env.get_G_Hm(r_probe=r_probe, r_source=r0, wavelength=wl)
            G_Ep_all = G_Ep.reshape(-1, 3)
            G_Hm_all = G_Hm.reshape(-1, 3)

            e_eval = e_eval.reshape(len(e_inc_extract), -1)
            h_eval = h_eval.reshape(len(e_inc_extract), -1)

            # inv. problem #1: probe fields + Green's tensors --> dipole moments
            G_Ep_i = torch.linalg.pinv(G_Ep_all)
            G_Hm_i = torch.linalg.pinv(G_Hm_all)
            p_eff = torch.matmul(G_Ep_i.unsqueeze(0), e_eval.unsqueeze(-1))[..., 0]
            m_eff = torch.matmul(G_Hm_i.unsqueeze(0), h_eval.unsqueeze(-1))[..., 0]

            # pseudo-inverse of all illuminations
            pinv_e0 = torch.linalg.pinv(e0_eval)
            pinv_h0 = torch.linalg.pinv(h0_eval)

            # optimum alphas to obtain dipole moments for each illumination
            alpha_pinv = torch.matmul(pinv_e0, p_eff)
            alpha_minv = torch.matmul(pinv_h0, m_eff)

            alpha_6x6[i_wl, :3, :3] = alpha_pinv
            alpha_6x6[i_wl, 3:, 3:] = alpha_minv

            # test residuals
            _test_residual_effective_polarizability(
                _struct, wl, env, alpha_pinv, e0_eval, p_eff, text_which_dp="ED"
            )
            _test_residual_effective_polarizability(
                _struct, wl, env, alpha_minv, h0_eval, m_eff, text_which_dp="MD"
            )

    if verbose:
        print("Done in {:.2f}s.".format(time.time() - t_start))
        print("---------------------------------------------")

    dict_pola_pseudo = dict(
        r0=r0,
        r0_MD=r0,
        r0_ED=r0,
        full_geometry=_struct.get_all_positions(),
        alpha_6x6=alpha_6x6,
        wavelengths=wavelengths,
        enclosing_radius=enclosing_radius,
        k0_spectrum=2 * torch.pi / wavelengths,
        environment=env,
    )

    return dict_pola_pseudo
